CUTLASS 3.8 Release (#2059)
* CUTLASS 3.8 Release * update * Update README.md * Revert "Update README.md" This reverts commit b353e36fe83e0815f99b44e46c0c95494c44726b. * update * update --------- Co-authored-by: Haicheng Wu <57973641+hwu36@users.noreply.github.com> Co-authored-by: Haicheng Wu <haichengw@nvidia.com>
This commit is contained in:
co-authored by
Haicheng Wu
Haicheng Wu
parent
9eb01fa0b0
commit
389e493055
Executable
+108
@@ -0,0 +1,108 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
|
||||
//
|
||||
|
||||
//
|
||||
|
||||
#pragma once
|
||||
|
||||
#include <cute/config.hpp>
|
||||
|
||||
|
||||
namespace cute {
|
||||
|
||||
|
||||
//
|
||||
// Cluster launch utility
|
||||
//
|
||||
CUTE_HOST
|
||||
bool
|
||||
initialize_preferred_cluster_launch(void const* const kernel_function,
|
||||
dim3 const& grid_dims,
|
||||
dim3 const& cluster_dims_preferred,
|
||||
dim3 const& cluster_dims_fallback)
|
||||
{
|
||||
//
|
||||
// Validate cluster_dims
|
||||
//
|
||||
|
||||
// Total number of cluster cannot be greater than 32 (hardware requirement)
|
||||
if (cluster_dims_preferred.x * cluster_dims_preferred.y * cluster_dims_preferred.z <= 0 ||
|
||||
cluster_dims_preferred.x * cluster_dims_preferred.y * cluster_dims_preferred.z > 32) {
|
||||
std::cout << "Invalid preferred cluster dimensions: Attempting to init preferred cluster (" << cluster_dims_preferred.x << "," << cluster_dims_preferred.y << "," << cluster_dims_preferred.z
|
||||
<< ") [" << (cluster_dims_preferred.x * cluster_dims_preferred.y * cluster_dims_preferred.z) << "] which must be within (0,32]." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
// Total number of cluster cannot be greater than 32 (hardware requirement)
|
||||
if (cluster_dims_fallback.x * cluster_dims_fallback.y * cluster_dims_fallback.z <= 0 ||
|
||||
cluster_dims_fallback.x * cluster_dims_fallback.y * cluster_dims_fallback.z > 32) {
|
||||
std::cout << "Invalid cluster dimensions: Attempting to init fallback cluster (" << cluster_dims_fallback.x << "," << cluster_dims_fallback.y << "," << cluster_dims_fallback.z
|
||||
<< ") [" << (cluster_dims_fallback.x * cluster_dims_fallback.y * cluster_dims_fallback.z) << "] which must be within (0,32]." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
// Total grid dimensions must be within (2^32, 2^16, 2^16)
|
||||
if (grid_dims.y > (1 << 16) || grid_dims.z > (1 << 16)) {
|
||||
std::cout << "Invalid grid dimensions: Attempting to init grid dimensions (" << grid_dims.x << "," << grid_dims.y << "," << grid_dims.z
|
||||
<< ") which must be within (2^32, 2^16, 2^16)." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
// grid_dims should be divisible by cluster_dims_preferred
|
||||
if (grid_dims.x % cluster_dims_preferred.x != 0 ||
|
||||
grid_dims.y % cluster_dims_preferred.y != 0 ||
|
||||
grid_dims.z % cluster_dims_preferred.z != 0) {
|
||||
std::cout << "Invalid grid dimensions: Preferred cluster (" << cluster_dims_preferred.x << "," << cluster_dims_preferred.y << "," << cluster_dims_preferred.z
|
||||
<< ") does not divide Grid (" << grid_dims.x << "," << grid_dims.y << "," << grid_dims.z << ")." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
// cluster_dims_preferred should be divisible by cluster_dims_fallback
|
||||
if (cluster_dims_preferred.x % cluster_dims_fallback.x != 0 ||
|
||||
cluster_dims_preferred.y % cluster_dims_fallback.y != 0 ||
|
||||
cluster_dims_preferred.z % cluster_dims_fallback.z != 0) {
|
||||
std::cout << "Invalid cluster dimensions: Fallback cluster (" << cluster_dims_fallback.x << "," << cluster_dims_fallback.y << "," << cluster_dims_fallback.z
|
||||
<< ") does not divide Preferred cluster (" << cluster_dims_preferred.x << "," << cluster_dims_preferred.y << "," << cluster_dims_preferred.z << ")." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
// Both cluster dimenions should have the same depth
|
||||
if (cluster_dims_preferred.z != cluster_dims_fallback.z) {
|
||||
std::cout << "Invalid cluster dimensions: Fallback cluster (" << cluster_dims_fallback.x << "," << cluster_dims_fallback.y << "," << cluster_dims_fallback.z
|
||||
<< ") and Preferred cluster (" << cluster_dims_preferred.x << "," << cluster_dims_preferred.y << "," << cluster_dims_preferred.z << ") does not have the same depth." << std::endl;
|
||||
return false;
|
||||
}
|
||||
|
||||
return true;
|
||||
}
|
||||
} // end namespace cute
|
||||
@@ -48,3 +48,42 @@
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
|
||||
#if (defined(CUTLASS_ARCH_MMA_SM100A_ENABLED))
|
||||
# define CUTE_ARCH_TMA_SM90_ENABLED
|
||||
# define CUTE_ARCH_DEVICE_MODIFIABLE_TMA_SM90_ENABLED
|
||||
# define CUTE_ARCH_STSM_SM90_ENABLED
|
||||
#endif
|
||||
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
# define CUTE_ARCH_TCGEN05_TF32_MMA_ENABLED
|
||||
# define CUTE_ARCH_TCGEN05_F16F32_MMA_ENABLED
|
||||
# define CUTE_ARCH_TCGEN05_MXF8F6F4_MMA_ENABLED
|
||||
# define CUTE_ARCH_TCGEN05_MXF4_MMA_ENABLED
|
||||
# define CUTE_ARCH_TCGEN05_MXF4NVF4_MMA_ENABLED
|
||||
#endif
|
||||
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
# define CUTE_ARCH_TCGEN05_S8_MMA_ENABLED
|
||||
#endif
|
||||
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
# define CUTE_ARCH_LDSM_SM100A_ENABLED
|
||||
# define CUTE_ARCH_STSM_SM100A_ENABLED
|
||||
#endif
|
||||
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
# define CUTE_ARCH_TCGEN05_TMEM_ENABLED
|
||||
#endif
|
||||
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
# define CUTE_ARCH_TMA_SM100_ENABLED
|
||||
#endif
|
||||
|
||||
// {add, mul, fma}.f32x2 PTX
|
||||
#if defined(CUTLASS_ARCH_MMA_SM100A_ENABLED)
|
||||
#define CUTE_ARCH_FLOAT2_MATH_ENABLED
|
||||
#endif
|
||||
|
||||
|
||||
|
||||
|
||||
File diff suppressed because it is too large
Load Diff
@@ -0,0 +1,664 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2020 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
|
||||
|
||||
|
||||
#pragma once
|
||||
|
||||
#include <cute/arch/config.hpp>
|
||||
|
||||
#include <cute/arch/copy.hpp>
|
||||
#include <cute/arch/copy_sm90.hpp>
|
||||
namespace cute
|
||||
{
|
||||
|
||||
constexpr uint32_t Sm100MmaPeerBitMask = 0xFEFFFFFF;
|
||||
constexpr uint64_t Sm100MemDescDefault = uint64_t(0x1000000000000000);
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
/// UTMA_LOAD : Initiates a TMA copy from global memory to shared memory
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_1D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy([[maybe_unused]] void const* desc_ptr, [[maybe_unused]] uint64_t* mbar_ptr, [[maybe_unused]] uint64_t cache_hint,
|
||||
[[maybe_unused]] void * smem_ptr,
|
||||
[[maybe_unused]] int32_t const& crd0)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.1d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3}], [%2], %4;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(crd0), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_2D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy([[maybe_unused]] void const* desc_ptr, [[maybe_unused]] uint64_t* mbar_ptr, [[maybe_unused]] uint64_t cache_hint,
|
||||
[[maybe_unused]] void * smem_ptr,
|
||||
[[maybe_unused]] int32_t const& crd0, int32_t const& crd1)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.2d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4}], [%2], %5;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(crd0), "r"(crd1), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_3D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy([[maybe_unused]] void const* desc_ptr, [[maybe_unused]] uint64_t* mbar_ptr, [[maybe_unused]] uint64_t cache_hint,
|
||||
[[maybe_unused]] void * smem_ptr,
|
||||
[[maybe_unused]] int32_t const& crd0, int32_t const& crd1, int32_t const& crd2)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.3d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5}], [%2], %6;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_4D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.4d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6}], [%2], %7;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "r"(crd3), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_5D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3, int32_t const& crd4)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.5d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6, %7}], [%2], %8;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "r"(crd3), "r"(crd4), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_1D::copy(desc_ptr, mbar_ptr, cache_hint, smem_ptr, crd0);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_2D::copy(desc_ptr, mbar_ptr, cache_hint, smem_ptr, crd0, crd1);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_3D::copy(desc_ptr, mbar_ptr, cache_hint, smem_ptr, crd0, crd1, crd2);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_4D::copy(desc_ptr, mbar_ptr, cache_hint, smem_ptr, crd0, crd1, crd2, crd3);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3, int32_t const& crd4)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_5D::copy(desc_ptr, mbar_ptr, cache_hint, smem_ptr, crd0, crd1, crd2, crd3, crd4);
|
||||
}
|
||||
|
||||
using PREFETCH = typename SM90_TMA_LOAD::PREFETCH;
|
||||
};
|
||||
|
||||
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
/// TMA_LOAD_MULTICAST: Initiates a TMA copy from global memory to shared memory
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_1D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.1d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%4}], [%2], %3, %5;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar), "h"(multicast_mask),
|
||||
"r"(crd0), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_2D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.2d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%4, %5}], [%2], %3, %6;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar), "h"(multicast_mask),
|
||||
"r"(crd0), "r"(crd1), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_3D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.3d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%4, %5, %6}], [%2], %3, %7;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar), "h"(multicast_mask),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_4D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.4d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%4, %5, %6, %7}], [%2], %3, %8;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar), "h"(multicast_mask),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "r"(crd3), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_5D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3, int32_t const& crd4)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.5d.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%4, %5, %6, %7, %8}], [%2], %3, %9;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar), "h"(multicast_mask),
|
||||
"r"(crd0), "r"(crd1), "r"(crd2), "r"(crd3), "r"(crd4), "l"(cache_hint)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM0_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_MULTICAST_1D::copy(desc_ptr, mbar_ptr, multicast_mask, cache_hint, smem_ptr, crd0);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_MULTICAST_2D::copy(desc_ptr, mbar_ptr, multicast_mask, cache_hint, smem_ptr, crd0, crd1);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_MULTICAST_3D::copy(desc_ptr, mbar_ptr, multicast_mask, cache_hint, smem_ptr, crd0, crd1, crd2);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_MULTICAST_4D::copy(desc_ptr, mbar_ptr, multicast_mask, cache_hint, smem_ptr, crd0, crd1, crd2, crd3);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask, uint64_t cache_hint,
|
||||
void * smem_ptr,
|
||||
int32_t const& crd0, int32_t const& crd1, int32_t const& crd2, int32_t const& crd3, int32_t const& crd4)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_MULTICAST_5D::copy(desc_ptr, mbar_ptr, multicast_mask, cache_hint, smem_ptr, crd0, crd1, crd2, crd3, crd4);
|
||||
}
|
||||
|
||||
using PREFETCH = typename SM90_TMA_LOAD::PREFETCH;
|
||||
};
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_3D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_n,
|
||||
uint16_t const& offset_w)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.3d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5}], [%2], {%6}, %7;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_n),
|
||||
"h"(offset_w), "l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_4D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.4d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6}], [%2], {%7, %8}, %9;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_h), "r"(coord_n),
|
||||
"h"(offset_w), "h"(offset_h), "l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_5D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_d, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h,
|
||||
uint16_t const& offset_d)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.5d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6, %7}], [%2], {%8, %9, %10}, %11;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_h), "r"(coord_d), "r"(coord_n),
|
||||
"h"(offset_w), "h"(offset_h), "h"(offset_d), "l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_n,
|
||||
uint16_t const& offset_w)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_3D::copy(desc_ptr, mbar_ptr, smem_ptr,
|
||||
coord_c, coord_w, coord_n,
|
||||
offset_w);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_4D::copy(desc_ptr, mbar_ptr, smem_ptr,
|
||||
coord_c, coord_w, coord_h, coord_n,
|
||||
offset_w, offset_h);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_d, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h,
|
||||
uint16_t const& offset_d)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_5D::copy(desc_ptr, mbar_ptr, smem_ptr,
|
||||
coord_c, coord_w, coord_h, coord_d, coord_n,
|
||||
offset_w, offset_h, offset_d);
|
||||
}
|
||||
|
||||
using PREFETCH = typename SM90_TMA_LOAD_IM2COL::PREFETCH;
|
||||
};
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_3D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_n,
|
||||
uint16_t const& offset_w)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.3d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5}], [%2], {%6}, %7, %8;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_n),
|
||||
"h"(offset_w),
|
||||
"h"(multicast_mask),
|
||||
"l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_4D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.4d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6}], [%2], {%7, %8}, %9, %10;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_h), "r"(coord_n),
|
||||
"h"(offset_w), "h"(offset_h),
|
||||
"h"(multicast_mask),
|
||||
"l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_5D
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_d, int32_t const& coord_n,
|
||||
uint16_t const& offset_w,
|
||||
uint16_t const& offset_h,
|
||||
uint16_t const& offset_d)
|
||||
{
|
||||
#if defined(CUTE_ARCH_TMA_SM100_ENABLED)
|
||||
uint64_t gmem_int_desc = reinterpret_cast<uint64_t>(desc_ptr);
|
||||
// Executed by both CTAs. Set peer bit to 0 so that the
|
||||
// transaction bytes will update CTA0's barrier.
|
||||
uint32_t smem_int_mbar = cast_smem_ptr_to_uint(mbar_ptr) & Sm100MmaPeerBitMask;
|
||||
uint32_t smem_int_ptr = cast_smem_ptr_to_uint(smem_ptr);
|
||||
asm volatile (
|
||||
"cp.async.bulk.tensor.5d.im2col.cta_group::2.shared::cluster.global.mbarrier::complete_tx::bytes.multicast::cluster.L2::cache_hint"
|
||||
" [%0], [%1, {%3, %4, %5, %6, %7}], [%2], {%8, %9, %10}, %11, %12;"
|
||||
:
|
||||
: "r"(smem_int_ptr), "l"(gmem_int_desc), "r"(smem_int_mbar),
|
||||
"r"(coord_c), "r"(coord_w), "r"(coord_h), "r"(coord_d), "r"(coord_n),
|
||||
"h"(offset_w), "h"(offset_h), "h"(offset_d),
|
||||
"h"(multicast_mask),
|
||||
"l"(Sm100MemDescDefault)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM100_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_MULTICAST
|
||||
{
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_n,
|
||||
uint16_t const& offset_w)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_3D::copy(desc_ptr, mbar_ptr, multicast_mask,
|
||||
smem_ptr,
|
||||
coord_c, coord_w, coord_n,
|
||||
offset_w);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_n,
|
||||
uint16_t const& offset_w, uint16_t const& offset_h)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_4D::copy(desc_ptr, mbar_ptr, multicast_mask,
|
||||
smem_ptr,
|
||||
coord_c, coord_w, coord_h, coord_n,
|
||||
offset_w, offset_h);
|
||||
}
|
||||
CUTE_HOST_DEVICE static void
|
||||
copy(void const* desc_ptr, uint64_t* mbar_ptr, uint16_t multicast_mask,
|
||||
void * smem_ptr,
|
||||
int32_t const& coord_c, int32_t const& coord_w, int32_t const& coord_h, int32_t const& coord_d, int32_t const& coord_n,
|
||||
uint16_t const& offset_w, uint16_t const& offset_h, uint16_t const& offset_d)
|
||||
{
|
||||
return SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_5D::copy(desc_ptr, mbar_ptr, multicast_mask,
|
||||
smem_ptr,
|
||||
coord_c, coord_w, coord_h, coord_d, coord_n,
|
||||
offset_w, offset_h, offset_d);
|
||||
}
|
||||
|
||||
using PREFETCH = typename SM90_TMA_LOAD_IM2COL::PREFETCH;
|
||||
};
|
||||
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
|
||||
} // end namespace cute
|
||||
@@ -140,6 +140,11 @@ enum class SmemSwizzleBits : uint8_t {
|
||||
|
||||
enum class SmemSwizzleBase : uint8_t {
|
||||
SWIZZLE_BASE_16B = 0,
|
||||
|
||||
SWIZZLE_BASE_32B = 1,
|
||||
SWIZZLE_BASE_32B_FLIP_8B = 2,
|
||||
SWIZZLE_BASE_64B = 3,
|
||||
|
||||
};
|
||||
|
||||
enum class OOBFill : uint8_t {
|
||||
@@ -184,6 +189,14 @@ enum class CacheHintSm90 : uint64_t {
|
||||
EVICT_LAST = 0x14F0000000000000,
|
||||
};
|
||||
|
||||
|
||||
enum class CacheHintSm100 : uint64_t {
|
||||
EVICT_NORMAL = 0x1000000000000000,
|
||||
EVICT_FIRST = 0x12F0000000000000,
|
||||
EVICT_LAST = 0x14F0000000000000,
|
||||
};
|
||||
|
||||
|
||||
#if (__CUDACC_VER_MAJOR__ >= 12)
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
@@ -195,6 +208,7 @@ to_CUtensorMapDataType() {
|
||||
if constexpr (is_same_v<T, uint8_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT8; } else
|
||||
if constexpr (is_same_v<T, float_e4m3_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT8; } else
|
||||
if constexpr (is_same_v<T, float_e5m2_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT8; } else
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float8_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT8;} else
|
||||
if constexpr (is_same_v<T, uint16_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT16; } else
|
||||
if constexpr (is_same_v<T, uint32_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT32; } else
|
||||
if constexpr (is_same_v<T, uint64_t>) { return CU_TENSOR_MAP_DATA_TYPE_UINT64; } else
|
||||
@@ -205,6 +219,18 @@ to_CUtensorMapDataType() {
|
||||
if constexpr (is_same_v<T, double>) { return CU_TENSOR_MAP_DATA_TYPE_FLOAT64; } else
|
||||
if constexpr (is_same_v<T, bfloat16_t>) { return CU_TENSOR_MAP_DATA_TYPE_BFLOAT16; } else
|
||||
if constexpr (is_same_v<T, tfloat32_t>) { return CU_TENSOR_MAP_DATA_TYPE_TFLOAT32; } else
|
||||
|
||||
if constexpr (is_same_v<T, float_e2m3_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, float_e3m2_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, float_e2m1_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U4_ALIGN8B;} else
|
||||
if constexpr (is_same_v<T, cutlass::detail::float_e2m1_unpacksmem_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U4_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, cutlass::detail::float_e2m3_unpacksmem_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, cutlass::detail::float_e3m2_unpacksmem_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, detail::type_erased_dynamic_float6_unpacksmem_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float6_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U6_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, detail::type_erased_dynamic_float4_unpacksmem_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U4_ALIGN16B;} else
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float4_t>) { return CU_TENSOR_MAP_DATA_TYPE_16U4_ALIGN8B; } else
|
||||
|
||||
{ static_assert(sizeof(T) < 0, "Unknown TMA Format!"); }
|
||||
}
|
||||
|
||||
@@ -221,9 +247,21 @@ to_CUtensorMapSwizzle(SmemSwizzleBits const& t, SmemSwizzleBase const& b) {
|
||||
case SmemSwizzleBits::B64:
|
||||
assert((b == SmemSwizzleBase::SWIZZLE_BASE_16B) && "Expected 16B swizzle base for 64B swizzle bits.");
|
||||
return CU_TENSOR_MAP_SWIZZLE_64B;
|
||||
#if (0)
|
||||
case SmemSwizzleBits::B128:
|
||||
assert((b == SmemSwizzleBase::SWIZZLE_BASE_16B) && "Expected 16B swizzle base for 128B swizzle bits.");
|
||||
return CU_TENSOR_MAP_SWIZZLE_128B;
|
||||
|
||||
#else
|
||||
case SmemSwizzleBits::B128:
|
||||
switch (b) {
|
||||
default: assert(false && "Unsupported pair of SmemSwizzleBits and SmemSwizzleBase!");
|
||||
case SmemSwizzleBase::SWIZZLE_BASE_16B: return CU_TENSOR_MAP_SWIZZLE_128B;
|
||||
case SmemSwizzleBase::SWIZZLE_BASE_32B: return CU_TENSOR_MAP_SWIZZLE_128B_ATOM_32B;
|
||||
case SmemSwizzleBase::SWIZZLE_BASE_64B: return CU_TENSOR_MAP_SWIZZLE_128B_ATOM_64B;
|
||||
}
|
||||
#endif
|
||||
|
||||
}
|
||||
}
|
||||
|
||||
|
||||
@@ -1157,6 +1157,17 @@ tma_store_arrive() {
|
||||
#endif
|
||||
}
|
||||
|
||||
|
||||
CUTE_HOST_DEVICE static void
|
||||
tma_desc_commit_group() {
|
||||
#if defined(CUTE_ARCH_TMA_SM90_ENABLED)
|
||||
asm volatile("cp.async.bulk.commit_group;");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM90_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
|
||||
|
||||
// Wait until at most Count committed TMA_STOREs are pending and all prior commits are complete
|
||||
template <int Count>
|
||||
CUTE_HOST_DEVICE static void
|
||||
@@ -1173,6 +1184,22 @@ tma_store_wait() {
|
||||
#endif
|
||||
}
|
||||
|
||||
|
||||
// Wait until all TMA descriptor previously issued are safe to be modified after tma_desc_commit_group()
|
||||
CUTE_HOST_DEVICE static void
|
||||
tma_desc_wait_group() {
|
||||
#if defined(CUTE_ARCH_TMA_SM90_ENABLED)
|
||||
asm volatile(
|
||||
"cp.async.bulk.wait_group.read %0;"
|
||||
:
|
||||
: "n"(0)
|
||||
: "memory");
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Trying to use tma without CUTE_ARCH_TMA_SM90_ENABLED.");
|
||||
#endif
|
||||
}
|
||||
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
/// TMA_REDUCE_ADD : Initiates a TMA reduce-add from shared memory to global memory
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
@@ -0,0 +1,42 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
//
|
||||
|
||||
//
|
||||
|
||||
#pragma once
|
||||
|
||||
#include <cute/arch/config.hpp>
|
||||
#include <cute/arch/mma.hpp>
|
||||
|
||||
namespace cute {
|
||||
|
||||
} // namespace cute
|
||||
@@ -0,0 +1,652 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
//
|
||||
|
||||
//
|
||||
|
||||
#pragma once
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
#include <cinttypes>
|
||||
#endif
|
||||
|
||||
#include <cute/arch/config.hpp>
|
||||
|
||||
#include <cute/arch/mma.hpp>
|
||||
|
||||
#include <cute/container/bit_field.hpp>
|
||||
#include <cute/container/array.hpp> // cute::array
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
namespace cute {
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
// UMMA Descriptor and utilities
|
||||
|
||||
// UMMA enums and utilities
|
||||
namespace UMMA
|
||||
{
|
||||
|
||||
enum class Major : uint8_t {
|
||||
K = 0,
|
||||
MN = 1
|
||||
};
|
||||
|
||||
enum class ScaleIn : uint8_t {
|
||||
One = 0,
|
||||
Neg = 1
|
||||
};
|
||||
|
||||
enum class ScaleOut : uint8_t {
|
||||
Zero = 0,
|
||||
One = 1
|
||||
};
|
||||
|
||||
enum class Saturate : uint8_t {
|
||||
False = 0,
|
||||
True = 1
|
||||
};
|
||||
|
||||
enum class LayoutType : uint8_t {
|
||||
SWIZZLE_NONE = 0,
|
||||
SWIZZLE_128B_BASE32B = 1,
|
||||
SWIZZLE_128B = 2,
|
||||
SWIZZLE_64B = 4,
|
||||
SWIZZLE_32B = 6
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(LayoutType const& t) {
|
||||
switch (t) {
|
||||
case LayoutType::SWIZZLE_NONE: return "SWIZZLE_NONE";
|
||||
case LayoutType::SWIZZLE_128B_BASE32B: return "SWIZZLE_128B_BASE32B";
|
||||
case LayoutType::SWIZZLE_128B: return "SWIZZLE_128B";
|
||||
case LayoutType::SWIZZLE_64B: return "SWIZZLE_64B";
|
||||
case LayoutType::SWIZZLE_32B: return "SWIZZLE_32B";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
union SmemDescriptor
|
||||
{
|
||||
uint64_t desc_ = 0;
|
||||
// Bitfield implementation avoids the need for shifts in assignment
|
||||
struct {
|
||||
// start_address, bit [0,14), 4LSB not included
|
||||
uint16_t start_address_ : 14, : 2; // 14 bits [0,14), 2 bits unused
|
||||
// leading dimension byte offset, bit [16,30), 4LSB not included
|
||||
uint16_t leading_byte_offset_ : 14, : 2; // 14 bits [0,14), 2 bits unused
|
||||
// stride dimension byte offset, bit [32,46), 4LSB not included
|
||||
uint16_t stride_byte_offset_ : 14, version_ : 2; // 14 bits [0,14), 2 bits [14,16)
|
||||
// base_offset, bit [49,52). leading_byte_offset_mode, bit [52,53).
|
||||
uint8_t : 1, base_offset_ : 3, lbo_mode_ : 1, : 3; // 1 bit unused, 3 bits [1,4), 1 bit [4,5), 3 bits unused
|
||||
// layout type, bit [61,64), SWIZZLE_NONE matrix descriptor = 0, SWIZZLE_128B matrix descriptor = 2, SWIZZLE_64B descriptor = 4, SWIZZLE_32B descriptor = 6, SWIZZLE_128B_BASE32B = 1, N/A = 3, N/A = 5, N/A = 7
|
||||
uint8_t : 5, layout_type_ : 3; // 6 bits unused, 3 bits [5,8)
|
||||
};
|
||||
// Seperate the field, as we may only update one part of desc
|
||||
struct {
|
||||
uint32_t lo;
|
||||
uint32_t hi;
|
||||
};
|
||||
|
||||
// Decay to a uint64_t
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
operator uint64_t() const noexcept { return desc_; }
|
||||
};
|
||||
|
||||
enum class F16F32Format : uint8_t {
|
||||
F16 = 0,
|
||||
BF16 = 1,
|
||||
TF32 = 2,
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(F16F32Format const& t) {
|
||||
switch (t) {
|
||||
case F16F32Format::F16: return "F16";
|
||||
case F16F32Format::BF16: return "BF16";
|
||||
case F16F32Format::TF32: return "TF32";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr F16F32Format to_F16F32Format() {
|
||||
if constexpr (is_same_v<T, half_t>) { return F16F32Format::F16; } else
|
||||
if constexpr (is_same_v<T, bfloat16_t>) { return F16F32Format::BF16; } else
|
||||
if constexpr (is_same_v<T, tfloat32_t>) { return F16F32Format::TF32; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for F16F32Format"); }
|
||||
}
|
||||
|
||||
enum class S8Format : uint8_t {
|
||||
UINT8 = 0,
|
||||
INT8 = 1,
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(S8Format const& t) {
|
||||
switch (t) {
|
||||
case S8Format::UINT8: return "UINT8";
|
||||
case S8Format::INT8: return "INT8";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr S8Format to_S8Format() {
|
||||
if constexpr (is_same_v<T, uint8_t>) { return S8Format::UINT8; } else
|
||||
if constexpr (is_same_v<T, int8_t>) { return S8Format::INT8; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for S8Format"); }
|
||||
}
|
||||
|
||||
enum class MXF8F6F4Format : uint8_t {
|
||||
E4M3 = 0,
|
||||
E5M2 = 1,
|
||||
E2M3 = 3,
|
||||
E3M2 = 4,
|
||||
E2M1 = 5,
|
||||
INVALID = 7 // an invalid datatype for runtime proxy type
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(MXF8F6F4Format const& t) {
|
||||
switch (t) {
|
||||
case MXF8F6F4Format::E4M3: return "E4M3";
|
||||
case MXF8F6F4Format::E5M2: return "E5M2";
|
||||
case MXF8F6F4Format::E2M3: return "E2M3";
|
||||
case MXF8F6F4Format::E3M2: return "E3M2";
|
||||
case MXF8F6F4Format::E2M1: return "E2M1";
|
||||
case MXF8F6F4Format::INVALID: return "INVALID";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr MXF8F6F4Format to_MXF8F6F4Format() {
|
||||
if constexpr (is_same_v<T, float_e4m3_t>) { return MXF8F6F4Format::E4M3; } else
|
||||
if constexpr (is_same_v<T, float_e5m2_t>) { return MXF8F6F4Format::E5M2; } else
|
||||
if constexpr (is_same_v<T, detail::float_e2m3_unpacksmem_t>) { return MXF8F6F4Format::E2M3; } else
|
||||
if constexpr (is_same_v<T, detail::float_e3m2_unpacksmem_t>) { return MXF8F6F4Format::E3M2; } else
|
||||
if constexpr (is_same_v<T, detail::float_e2m1_unpacksmem_t>) { return MXF8F6F4Format::E2M1; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for MXF8F6F4Format"); }
|
||||
}
|
||||
|
||||
enum class MXF4Format : uint8_t {
|
||||
E2M1 = 1,
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(MXF4Format const& t) {
|
||||
switch (t) {
|
||||
case MXF4Format::E2M1: return "E2M1";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr MXF4Format to_MXF4Format() {
|
||||
if constexpr (is_same_v<T, float_e2m1_t>) { return MXF4Format::E2M1; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for MXF4Format"); }
|
||||
}
|
||||
|
||||
enum class ScaleFormat : uint8_t {
|
||||
UE4M3 = 0,
|
||||
UE8M0 = 1,
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(ScaleFormat const& t) {
|
||||
switch (t) {
|
||||
case ScaleFormat::UE4M3: return "UE4M3";
|
||||
case ScaleFormat::UE8M0: return "UE8M0";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr ScaleFormat to_ScaleFormat() {
|
||||
if constexpr (is_same_v<T, float_ue4m3_t>) { return ScaleFormat::UE4M3; } else
|
||||
if constexpr (is_same_v<T, float_ue8m0_t>) { return ScaleFormat::UE8M0; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for ScaleFormat"); }
|
||||
}
|
||||
|
||||
enum class CFormat : uint8_t {
|
||||
F16 = 0,
|
||||
F32 = 1,
|
||||
S32 = 2,
|
||||
};
|
||||
|
||||
CUTE_HOST_DEVICE char const* to_string(CFormat const& t) {
|
||||
switch (t) {
|
||||
case CFormat::F16: return "F16";
|
||||
case CFormat::F32: return "F32";
|
||||
case CFormat::S32: return "S32";
|
||||
}
|
||||
return nullptr;
|
||||
}
|
||||
|
||||
enum class MaxShift : uint8_t {
|
||||
NoShift = 0,
|
||||
MaxShift8 = 1,
|
||||
MaxShift16 = 2,
|
||||
MaxShift32 = 3
|
||||
};
|
||||
|
||||
enum class BMatrixBufferId : uint8_t {
|
||||
Zero = 0u,
|
||||
One = 1u,
|
||||
Two = 2u,
|
||||
Three = 3u
|
||||
};
|
||||
|
||||
enum class BMatrixBufferReuse : uint8_t {
|
||||
Keep = 1u,
|
||||
Reuse = 2u,
|
||||
ReuseAndKeep = 3u
|
||||
};
|
||||
|
||||
// using MaskAndShiftB = uint32_t[2];
|
||||
union MaskAndShiftB
|
||||
{
|
||||
uint32_t uri[2];
|
||||
|
||||
struct {
|
||||
// Bitfield implementation avoids the need for shifts in assignment
|
||||
uint8_t start_count_ [4]; // bit [ 0:32) : 8 bits each. Specifies the start count for mask generation.
|
||||
uint32_t first_span_ : 4, // bit [32:36) : 1 bit each. 0 = start where B is used. 1 = start with where B is skipped(0 value is used).
|
||||
: 3, //
|
||||
nzm_ : 1, // bit [39:40) : 0 = Enable the mask. 1 = Disable the mask.
|
||||
skip_span_ : 8, // bit [40:48) : Count-1 (zero encoded in this field specifies use span of 1) of consecutive columns where 0 value is used.
|
||||
use_span_ : 8, // bit [48:55) : Count-1 (zero encoded in this field specifies use span of 1) of consecutive columns where B matrix data is used.
|
||||
shift_ : 6, // bit [56:62) : Shift value for B matrix data.
|
||||
: 2;
|
||||
};
|
||||
};
|
||||
|
||||
template <typename ShapeType, int FLT_S, int CTA_M, int CTA_N>
|
||||
CUTE_HOST_DEVICE constexpr auto
|
||||
make_column_zero_mask(ShapeType conv_q, int32_t cta_coord_q, int32_t num_pixels_skip_left) {
|
||||
|
||||
static_assert(cute::is_same_v<ShapeType, cutlass::FastDivmod> || cute::is_integral<ShapeType>::value);
|
||||
|
||||
cute::array<MaskAndShiftB, FLT_S> column_zero_masks{};
|
||||
|
||||
static_assert(FLT_S == 3, "Filter size not supported.");
|
||||
constexpr int MAX_USE_SPAN_COUNT = 256;
|
||||
constexpr int MAX_SKIP_SPAN_COUNT = 256;
|
||||
|
||||
// conv_q_int used for non-divmod case (add/minus/..)
|
||||
// conv_q used for divmod case (div/mod/...)
|
||||
int32_t conv_q_int = int(conv_q);
|
||||
auto [_, cta_q] = divmod(cta_coord_q * CTA_N, conv_q);
|
||||
|
||||
int step_q = CTA_M == 128 ? CTA_N / 1
|
||||
: CTA_M == 64 ? CTA_N / 2
|
||||
: CTA_M == 32 ? CTA_N / 4
|
||||
: 0;
|
||||
|
||||
for (int mask_iter = 0; mask_iter < int(CTA_N / step_q); ++mask_iter) {
|
||||
|
||||
for (int s_iter = 0; s_iter < FLT_S; s_iter += 1) {
|
||||
|
||||
int32_t skip_span{0}, use_span{0}, nzm{1}, first_span{0}, start_count{0}, shift{0};
|
||||
|
||||
shift = s_iter;
|
||||
|
||||
// Examples for CZM setting
|
||||
// CASE0: (skip_span_ < 0)
|
||||
// | padding |<- conv_q ->|
|
||||
// |skip_span_|<- use_span ->|skip_span_|
|
||||
// -skip_span 0 ^cta_q conv_q-1
|
||||
// 0 ^index
|
||||
//
|
||||
// CASE1: (skip_span_ > 0)
|
||||
// |<- conv_q ->|
|
||||
// |skip_span_|<- use_span ->|skip_span_|
|
||||
// 0 ^cta_q conv_q-1
|
||||
// 0 ^index
|
||||
//
|
||||
// line 0 an input vector from 0 to conv_q with the padding
|
||||
// line 1 shows the different spans we need to skip or load
|
||||
// lines 2-3 show the different coordinates of different boundaries.
|
||||
// CTQ_q is the coordinate of the present cta.
|
||||
|
||||
int32_t skip_span_ = num_pixels_skip_left - shift;
|
||||
int32_t index{0};
|
||||
if (skip_span_ > 0) {
|
||||
auto [_, index_mod] = divmod(cta_q, conv_q);
|
||||
index = index_mod;
|
||||
} else if (skip_span_ < 0) {
|
||||
auto [_, index_mod] = divmod((cta_q - skip_span_), conv_q);
|
||||
index = index_mod;
|
||||
} else {
|
||||
nzm = 0;
|
||||
}
|
||||
skip_span = cute::max(cute::abs(skip_span_), 1);
|
||||
use_span = cute::min(conv_q_int - static_cast<int32_t>(skip_span), MAX_USE_SPAN_COUNT);
|
||||
if (use_span > 0) {
|
||||
first_span = index >= skip_span ? 0 : 1;
|
||||
if ((first_span == 0) && (index + CTA_N < conv_q_int + skip_span)) {
|
||||
nzm = 0;
|
||||
} else {
|
||||
start_count = first_span == 0 ? (use_span - (conv_q_int - index)) : index;
|
||||
}
|
||||
} else {
|
||||
skip_span = MAX_SKIP_SPAN_COUNT;
|
||||
use_span = 1;
|
||||
first_span = 1;
|
||||
start_count = 0;
|
||||
}
|
||||
|
||||
column_zero_masks[s_iter].start_count_[mask_iter] = start_count;
|
||||
column_zero_masks[s_iter].first_span_ |= first_span << mask_iter;
|
||||
column_zero_masks[s_iter].nzm_ |= nzm;
|
||||
column_zero_masks[s_iter].skip_span_ = skip_span - 1;
|
||||
column_zero_masks[s_iter].use_span_ = use_span - 1;
|
||||
column_zero_masks[s_iter].shift_ = shift;
|
||||
|
||||
}
|
||||
|
||||
cta_q += step_q;
|
||||
}
|
||||
|
||||
return column_zero_masks;
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr auto to_UMMAFormat() {
|
||||
if constexpr (is_same_v<T, half_t>) { return F16F32Format::F16; } else
|
||||
if constexpr (is_same_v<T, bfloat16_t>) { return F16F32Format::BF16; } else
|
||||
if constexpr (is_same_v<T, tfloat32_t>) { return F16F32Format::TF32; } else
|
||||
if constexpr (is_same_v<T, uint8_t>) { return S8Format::UINT8; } else
|
||||
if constexpr (is_same_v<T, int8_t>) { return S8Format::INT8; } else
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float8_t>) {return MXF8F6F4Format::INVALID; } else
|
||||
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float6_t>) {return MXF8F6F4Format::INVALID; } else
|
||||
if constexpr (is_same_v<T, type_erased_dynamic_float4_t>) {return MXF8F6F4Format::INVALID; } else
|
||||
if constexpr (is_same_v<T, detail::type_erased_dynamic_float4_unpacksmem_t>) {return MXF8F6F4Format::INVALID; } else
|
||||
|
||||
if constexpr (is_same_v<T, float_e4m3_t>) { return MXF8F6F4Format::E4M3; } else
|
||||
if constexpr (is_same_v<T, float_e5m2_t>) { return MXF8F6F4Format::E5M2; } else
|
||||
|
||||
if constexpr (is_same_v<T, detail::type_erased_dynamic_float6_unpacksmem_t>) {return MXF8F6F4Format::INVALID; } else
|
||||
if constexpr (is_same_v<T, detail::float_e2m3_unpacksmem_t>) { return MXF8F6F4Format::E2M3; } else
|
||||
if constexpr (is_same_v<T, detail::float_e3m2_unpacksmem_t>) { return MXF8F6F4Format::E3M2; } else
|
||||
if constexpr (is_same_v<T, float_e2m3_t>) { return MXF8F6F4Format::E2M3; } else
|
||||
if constexpr (is_same_v<T, float_e3m2_t>) { return MXF8F6F4Format::E3M2; } else
|
||||
if constexpr (is_same_v<T, detail::float_e2m1_unpacksmem_t>) { return MXF8F6F4Format::E2M1; } else
|
||||
if constexpr (is_same_v<T, float_e2m1_t>) { return MXF4Format::E2M1; } else
|
||||
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for UMMAFormat"); }
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr CFormat to_CFormat() {
|
||||
if constexpr (is_same_v<T, half_t>) { return CFormat::F16; } else
|
||||
if constexpr (is_same_v<T, float>) { return CFormat::F32; } else
|
||||
if constexpr (is_same_v<T, int32_t>) { return CFormat::S32; } else
|
||||
{ static_assert(sizeof(T) == 0, "Unknown type for CFormat"); }
|
||||
}
|
||||
|
||||
union InstrDescriptor
|
||||
{
|
||||
uint32_t desc_;
|
||||
|
||||
struct {
|
||||
// Bitfield implementation avoids the need for shifts in assignment
|
||||
uint16_t sparse_id2_ : 2, // bit [ 0, 2) : Sparse meta data id2
|
||||
sparse_flag_ : 1, // bit [ 2, 3) : 0 = dense. 1 = sparse. 1 value valid only for F32F16/S8/MXF8F6F4
|
||||
saturate_ : 1, // bit [ 3, 4) : 0 = no saturate. 1 = saturate. 1 value valid only for S8
|
||||
c_format_ : 2, // bit [ 4, 6) : 0 = F16. 1 = F32, 2 = S32
|
||||
: 1, //
|
||||
a_format_ : 3, // bit [ 7,10) : MXF8F6F4Format:0 = E4M3, 1 = E5M2, 3 = E2M3, 4 = E3M2, 5 = E2M1. F32F16Format: 0 = F16, 1 = BF16, 2 = TF32. S8: 0 unsigned 8 bit, 1 signed 8 bit. Boolean MMA: 0 Boolean
|
||||
b_format_ : 3, // bit [10,13) : MXF8F6F4Format:0 = E4M3, 1 = E5M2, 3 = E2M3, 4 = E3M2, 5 = E2M1. F32F16Format: 0 = F16, 1 = BF16, 2 = TF32. S8: 0 unsigned 8 bit, 1 signed 8 bit. Boolean MMA: 0 Boolean
|
||||
a_negate_ : 1, // bit [13,14) : 0 = no negate. 1 = negate. 1 value valid only for F32F16Format and MXF8F6F4Format
|
||||
b_negate_ : 1, // bit [14,15) : 0 = no negate. 1 = negate. 1 value valid only for F32F16Format and MXF8F6F4Format
|
||||
a_major_ : 1; // bit [15,16) : 0 = K-major. 1 = MN-major. Major value of 1 is only valid for E4M3, E5M2, INT8 (signed and unsigned), F16, BF16 and TF32 source formats
|
||||
uint16_t b_major_ : 1, // bit [16,17) : 0 = K-major. 1 = MN-major. Major value of 1 is only valid for E4M3, E5M2, INT8 (signed and unsigned), F16, BF16 and TF32 source formats
|
||||
n_dim_ : 6, // bit [17,23) : 3 LSBs not included. Valid values range from 1 (N=8) to 32 (N=256). All values are not valid for all instruction formats
|
||||
: 1, //
|
||||
m_dim_ : 5, // bit [24,29) : 4 LSBs not included. Valid values are: 4 (M=64), 8 (M=128), 16 (M=256)
|
||||
: 1, //
|
||||
max_shift_ : 2; // bit [30,32) : Maximum shift for WS instruction. Encoded as follows: 0 = no shift, 1 = maximum shift of 8, 2 = maximum shift of 16, 3 = maximum shift of 32.
|
||||
};
|
||||
|
||||
// Decay to a uint32_t
|
||||
CUTE_HOST_DEVICE constexpr explicit
|
||||
operator uint32_t() const noexcept { return desc_; }
|
||||
};
|
||||
|
||||
union InstrDescriptorBlockScaled
|
||||
{
|
||||
uint32_t desc_;
|
||||
|
||||
struct {
|
||||
// Bitfield implementation avoids the need for shifts in assignment
|
||||
uint16_t sparse_id2_ : 2, // bit [ 0, 2) : Sparse meta data id2
|
||||
sparse_flag_ : 1, // bit [ 2, 3) : 0 = dense. 1 = sparse. 1 value valid only for F32F16/S8/MXF8F6F4
|
||||
: 1, //
|
||||
b_sf_id_ : 2, // bit [ 4, 6) : Matrix B Scale Factor ID
|
||||
: 1, //
|
||||
a_format_ : 3, // bit [ 7, 9) : MXF8F6F4Format:0 = E4M3, 1 = E5M2, 3 = E2M3, 4 = E3M2, 5 = E2M1. F32F16Format: 0 = F16, 1 = BF16, 2 = TF32. S8: 0 unsigned 8 bit, 1 signed 8 bit. BMMA: 0 Boolean
|
||||
b_format_ : 3, // bit [10,12) : MXF8F6F4Format:0 = E4M3, 1 = E5M2, 3 = E2M3, 4 = E3M2, 5 = E2M1. F32F16Format: 0 = F16, 1 = BF16, 2 = TF32. S8: 0 unsigned 8 bit, 1 signed 8 bit. BMMA: 0 Boolean
|
||||
a_negate_ : 1, // bit [13,14) : 0 = no negate. 1 = negate. 1 value valid only for F32F16Format and MXF8F6F4Format
|
||||
b_negate_ : 1, // bit [14,15) : 0 = no negate. 1 = negate. 1 value valid only for F32F16Format and MXF8F6F4Format
|
||||
a_major_ : 1; // bit [15,16) : 0 = K-major. 1 = MN-major. Major value of 1 is only valid for E4M3, E5M2, INT8 (signed and unsigned), F16, BF16 and TF32 source formats
|
||||
uint16_t b_major_ : 1, // bit [16,17) : 0 = K-major. 1 = MN-major. Major value of 1 is only valid for E4M3, E5M2, INT8 (signed and unsigned), F16, BF16 and TF32 source formats
|
||||
n_dim_ : 6, // bit [17,23) : 3 LSBs not included. Valid values range from 1 (N=8) to 32 (N=256). All values are not valid for all instruction formats
|
||||
scale_format_ : 1, // bit [23,24) : 0=E4M3, 1=E8M0
|
||||
m_dim_ : 5, // bit [24,29) : 4 LSBs not included. Valid values are: 4 (M=64), 8 (M=128), 16 (M=256)
|
||||
a_sf_id_ : 2, // bit [29,31) : Matrix A Scale Factor ID
|
||||
: 1; //
|
||||
};
|
||||
|
||||
// Decay to a uint32_t
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
operator uint32_t() const noexcept { return desc_; }
|
||||
};
|
||||
|
||||
template <class a_type, class b_type, class c_type,
|
||||
int M, int N, UMMA::Major a_major, UMMA::Major b_major,
|
||||
UMMA::ScaleIn a_neg = UMMA::ScaleIn::One, UMMA::ScaleIn b_neg = UMMA::ScaleIn::One,
|
||||
UMMA::Saturate c_sat = UMMA::Saturate::False,
|
||||
bool is_sparse = false,
|
||||
UMMA::MaxShift max_shift = UMMA::MaxShift::NoShift>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
UMMA::InstrDescriptor
|
||||
make_instr_desc()
|
||||
{
|
||||
UMMA::InstrDescriptor desc_i = {};
|
||||
|
||||
desc_i.a_format_ = uint8_t(UMMA::to_UMMAFormat<a_type>());
|
||||
desc_i.b_format_ = uint8_t(UMMA::to_UMMAFormat<b_type>());
|
||||
desc_i.c_format_ = uint8_t(UMMA::to_CFormat<c_type>());
|
||||
|
||||
desc_i.m_dim_ = (M >> 4);
|
||||
desc_i.n_dim_ = (N >> 3);
|
||||
|
||||
desc_i.a_major_ = uint8_t(a_major);
|
||||
desc_i.b_major_ = uint8_t(b_major);
|
||||
|
||||
desc_i.a_negate_ = uint8_t(a_neg);
|
||||
desc_i.b_negate_ = uint8_t(b_neg);
|
||||
desc_i.saturate_ = uint8_t(c_sat);
|
||||
|
||||
desc_i.sparse_flag_ = is_sparse; // 1 = Sparse
|
||||
desc_i.sparse_id2_ = 0;
|
||||
|
||||
desc_i.max_shift_ = uint8_t(max_shift);
|
||||
|
||||
return desc_i;
|
||||
}
|
||||
|
||||
template <class a_type, class b_type, class c_type,
|
||||
int M, int N, UMMA::Major a_major, UMMA::Major b_major,
|
||||
UMMA::ScaleIn a_neg = UMMA::ScaleIn::One, UMMA::ScaleIn b_neg = UMMA::ScaleIn::One,
|
||||
UMMA::Saturate c_sat = UMMA::Saturate::False,
|
||||
bool is_sparse = false,
|
||||
UMMA::MaxShift max_shift = UMMA::MaxShift::NoShift>
|
||||
CUTE_HOST_DEVICE
|
||||
constexpr uint64_t
|
||||
make_runtime_instr_desc(uint16_t sparse_id2 = 0u, uint32_t tmem_e = 0u) {
|
||||
UMMA::InstrDescriptor desc_i = UMMA::make_instr_desc<
|
||||
a_type, b_type, c_type, M, N, a_major, b_major, a_neg, b_neg, c_sat, is_sparse,
|
||||
max_shift>();
|
||||
|
||||
if constexpr (is_sparse) {
|
||||
desc_i.sparse_id2_ = sparse_id2;
|
||||
}
|
||||
else {
|
||||
assert(sparse_id2 == 0u);
|
||||
}
|
||||
// In current compiler exposure, idescE is a uint64_t. It should contain:
|
||||
// - Lower 32b URe: Specifies the tmem address that stores the sparse metadata.
|
||||
// Only needed for Sparse MMA instructions. Otherwise, ignored.
|
||||
// - Upper 32b URh: Specifies the instruction descriptor.
|
||||
uint64_t idescE = (static_cast<uint64_t>(static_cast<uint32_t>(desc_i)) << 32);
|
||||
|
||||
return idescE;
|
||||
}
|
||||
|
||||
template <bool is_sparse = false>
|
||||
CUTE_HOST_DEVICE
|
||||
constexpr uint64_t
|
||||
make_runtime_instr_desc(UMMA::InstrDescriptor desc_i, uint16_t sparse_id2 = 0u, uint32_t tmem_e = 0u)
|
||||
{
|
||||
if constexpr (is_sparse) {
|
||||
desc_i.sparse_id2_ = sparse_id2;
|
||||
}
|
||||
else {
|
||||
assert(sparse_id2 == 0u);
|
||||
}
|
||||
// In current compiler exposure, idescE is a uint64_t. It should contain:
|
||||
// - Lower 32b URe: Specifies the tmem address that stores the sparse metadata.
|
||||
// Only needed for Sparse MMA instructions. Otherwise, ignored.
|
||||
// - Upper 32b URh: Specifies the instruction descriptor.
|
||||
uint64_t idescE = (static_cast<uint64_t>(static_cast<uint32_t>(desc_i)) << 32);
|
||||
|
||||
return idescE;
|
||||
}
|
||||
|
||||
template <class a_type, class b_type, class c_type, class sf_type,
|
||||
int M, int N, UMMA::Major a_major, UMMA::Major b_major,
|
||||
UMMA::ScaleIn a_neg = UMMA::ScaleIn::One, UMMA::ScaleIn b_neg = UMMA::ScaleIn::One,
|
||||
bool is_sparse = false>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
UMMA::InstrDescriptorBlockScaled
|
||||
make_instr_desc_block_scaled()
|
||||
{
|
||||
UMMA::InstrDescriptorBlockScaled desc_i = {};
|
||||
|
||||
desc_i.a_format_ = uint8_t(UMMA::to_UMMAFormat<a_type>());
|
||||
desc_i.b_format_ = uint8_t(UMMA::to_UMMAFormat<b_type>());
|
||||
|
||||
desc_i.scale_format_ = uint8_t(UMMA::to_ScaleFormat<sf_type>());
|
||||
desc_i.a_sf_id_ = 0;
|
||||
desc_i.b_sf_id_ = 0;
|
||||
|
||||
desc_i.m_dim_ = (M >> 4);
|
||||
desc_i.n_dim_ = (N >> 3);
|
||||
|
||||
desc_i.a_major_ = uint8_t(a_major);
|
||||
desc_i.b_major_ = uint8_t(b_major);
|
||||
|
||||
desc_i.a_negate_ = uint8_t(a_neg);
|
||||
desc_i.b_negate_ = uint8_t(b_neg);
|
||||
desc_i.sparse_flag_ = is_sparse; // 1 = Sparse
|
||||
desc_i.sparse_id2_ = 0;
|
||||
|
||||
// Below would bring some warnings.
|
||||
#if defined(__GNUC__)
|
||||
# pragma GCC diagnostic ignored "-Wconversion"
|
||||
#endif
|
||||
return desc_i;
|
||||
}
|
||||
|
||||
template <class a_type, class b_type, class c_type, class sf_type,
|
||||
int M, int N, UMMA::Major a_major, UMMA::Major b_major,
|
||||
UMMA::ScaleIn a_neg = UMMA::ScaleIn::One, UMMA::ScaleIn b_neg = UMMA::ScaleIn::One,
|
||||
bool is_sparse = false>
|
||||
CUTE_HOST_DEVICE
|
||||
constexpr uint64_t
|
||||
make_runtime_instr_desc_block_scaled(uint32_t const tmem_sfa_addr, uint32_t const tmem_sfb_addr,
|
||||
uint16_t const sparse_id2 = 0u, uint32_t const tmem_e = 0u)
|
||||
{
|
||||
UMMA::InstrDescriptorBlockScaled desc_i = UMMA::make_instr_desc_block_scaled<
|
||||
a_type, b_type, c_type, sf_type, M, N,
|
||||
a_major, b_major,
|
||||
a_neg, b_neg,
|
||||
is_sparse>();
|
||||
|
||||
// The first 2-bits of TMEM address includes byte address.
|
||||
desc_i.a_sf_id_ = (tmem_sfa_addr & 0xC0000000) >> 30;
|
||||
desc_i.b_sf_id_ = (tmem_sfb_addr & 0xC0000000) >> 30;
|
||||
|
||||
if constexpr (is_sparse) {
|
||||
desc_i.sparse_id2_ = sparse_id2;
|
||||
}
|
||||
else {
|
||||
assert(sparse_id2 == 0u);
|
||||
}
|
||||
|
||||
// In current compiler exposure, idescE is a uint64_t. It should contain:
|
||||
// - Lower 32b URe: Specifies the tmem address that stores the sparse metadata.
|
||||
// Only needed for Sparse MMA instructions. Otherwise, ignored.
|
||||
// - Upper 32b URh: Specifies the instruction descriptor.
|
||||
uint64_t idescE = (static_cast<uint64_t>(static_cast<uint32_t>(desc_i)) << 32);
|
||||
|
||||
return idescE;
|
||||
}
|
||||
|
||||
template <bool is_sparse = false>
|
||||
CUTE_HOST_DEVICE
|
||||
constexpr uint64_t
|
||||
make_runtime_instr_desc_block_scaled(UMMA::InstrDescriptorBlockScaled desc_i,
|
||||
uint32_t const tmem_sfa_addr, uint32_t const tmem_sfb_addr,
|
||||
uint16_t const sparse_id2 = 0u, uint32_t const tmem_e = 0u)
|
||||
{
|
||||
// The first 2-bits of TMEM address includes byte address.
|
||||
desc_i.a_sf_id_ = (tmem_sfa_addr & 0xC0000000) >> 30;
|
||||
desc_i.b_sf_id_ = (tmem_sfb_addr & 0xC0000000) >> 30;
|
||||
|
||||
if constexpr (is_sparse) {
|
||||
desc_i.sparse_id2_ = sparse_id2;
|
||||
}
|
||||
else {
|
||||
assert(sparse_id2 == 0u);
|
||||
}
|
||||
|
||||
// In current compiler exposure, idescE is a uint64_t. It should contain:
|
||||
// - Lower 32b URe: Specifies the tmem address that stores the sparse metadata.
|
||||
// Only needed for Sparse MMA instructions. Otherwise, ignored.
|
||||
// - Upper 32b URh: Specifies the instruction descriptor.
|
||||
uint64_t idescE = (static_cast<uint64_t>(static_cast<uint32_t>(desc_i)) << 32);
|
||||
|
||||
return idescE;
|
||||
}
|
||||
|
||||
} // end namespace UMMA
|
||||
} // namespace cute
|
||||
File diff suppressed because it is too large
Load Diff
@@ -0,0 +1,96 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2024 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
//
|
||||
|
||||
//
|
||||
|
||||
#pragma once
|
||||
|
||||
#include <cute/config.hpp>
|
||||
#include <cute/arch/config.hpp>
|
||||
#include <cute/numeric/real.hpp>
|
||||
|
||||
namespace cute {
|
||||
|
||||
CUTE_HOST_DEVICE
|
||||
void
|
||||
add(float2 & c,
|
||||
float2 const& a,
|
||||
float2 const& b)
|
||||
{
|
||||
#if defined(CUTE_ARCH_FLOAT2_MATH_ENABLED)
|
||||
asm volatile("add.f32x2 %0, %1, %2;\n"
|
||||
: "=l"(reinterpret_cast<uint64_t &>(c))
|
||||
: "l"(reinterpret_cast<uint64_t const&>(a)),
|
||||
"l"(reinterpret_cast<uint64_t const&>(b)));
|
||||
#else
|
||||
add(c.x, a.x, b.x);
|
||||
add(c.y, a.y, b.y);
|
||||
#endif
|
||||
}
|
||||
|
||||
CUTE_HOST_DEVICE
|
||||
void
|
||||
mul(float2 & c,
|
||||
float2 const& a,
|
||||
float2 const& b)
|
||||
{
|
||||
#if defined(CUTE_ARCH_FLOAT2_MATH_ENABLED)
|
||||
asm volatile("mul.f32x2 %0, %1, %2;\n"
|
||||
: "=l"(reinterpret_cast<uint64_t &>(c))
|
||||
: "l"(reinterpret_cast<uint64_t const&>(a)),
|
||||
"l"(reinterpret_cast<uint64_t const&>(b)));
|
||||
#else
|
||||
mul(c.x, a.x, b.x);
|
||||
mul(c.y, a.y, b.y);
|
||||
#endif
|
||||
}
|
||||
|
||||
CUTE_HOST_DEVICE
|
||||
void
|
||||
fma(float2 & d,
|
||||
float2 const& a,
|
||||
float2 const& b,
|
||||
float2 const& c)
|
||||
{
|
||||
#if defined(CUTE_ARCH_FLOAT2_MATH_ENABLED)
|
||||
asm volatile("fma.rn.f32x2 %0, %1, %2, %3;\n"
|
||||
: "=l"(reinterpret_cast<uint64_t &>(d))
|
||||
: "l"(reinterpret_cast<uint64_t const&>(a)),
|
||||
"l"(reinterpret_cast<uint64_t const&>(b)),
|
||||
"l"(reinterpret_cast<uint64_t const&>(c)));
|
||||
#else
|
||||
fma(d.x, a.x, b.x, c.x);
|
||||
fma(d.y, a.y, b.y, c.y);
|
||||
#endif
|
||||
}
|
||||
|
||||
} // namespace cute
|
||||
@@ -0,0 +1,168 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
//
|
||||
|
||||
//
|
||||
#pragma once
|
||||
|
||||
#include <cute/arch/config.hpp>
|
||||
#include <cute/arch/cluster_sm90.hpp>
|
||||
#include <cute/atom/copy_traits_sm100.hpp>
|
||||
|
||||
#include <cutlass/pipeline/sm90_pipeline.hpp>
|
||||
|
||||
namespace cute::TMEM {
|
||||
|
||||
// All operations of this class require that only a single warp uniformly participates
|
||||
class Allocator1Sm {
|
||||
public:
|
||||
static constexpr int ColumnsPerAllocationSlice = 32;
|
||||
static constexpr int Sm100TmemCapacityColumns = 512;
|
||||
|
||||
__device__ Allocator1Sm() { }
|
||||
|
||||
/**
|
||||
* Performs a non-blocking allocation of TMEM.
|
||||
* @param num_columns Number of columns being freed. Must be 32 <= num_columns <= 512 and power of 2.
|
||||
* @param dst_ptr Pointer to shared memory to which to write the result tmem pointer to.
|
||||
* @pre Must be issued by a single fully active warp of the CTA.
|
||||
* @pre Must never be issued by more than one warp at the same time.
|
||||
* @pre For repeated allocations, the same warp must be used to issue all allocations.
|
||||
**/
|
||||
__device__ void
|
||||
allocate(int num_columns, uint32_t* dst_ptr) {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
uint32_t dst_intptr = cute::cast_smem_ptr_to_uint(dst_ptr);
|
||||
asm volatile(
|
||||
"tcgen05.alloc.cta_group::1.sync.aligned.shared::cta.b32 [%0], %1;"
|
||||
:
|
||||
: "r"(dst_intptr), "r"(num_columns));
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
|
||||
__device__
|
||||
void
|
||||
free(uint32_t tmem_ptr, int num_columns) {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
asm volatile(
|
||||
"{\n\t"
|
||||
"tcgen05.dealloc.cta_group::1.sync.aligned.b32 %0, %1; \n\t"
|
||||
"}"
|
||||
:
|
||||
: "r"(tmem_ptr), "r"(num_columns));
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
|
||||
__device__ void
|
||||
release_allocation_lock() {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
asm volatile("tcgen05.relinquish_alloc_permit.cta_group::1.sync.aligned;" ::);
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
///////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
class Allocator2Sm {
|
||||
public:
|
||||
static constexpr int ColumnsPerAllocationSlice = 32;
|
||||
static constexpr int Sm100TmemCapacityColumns = 512;
|
||||
|
||||
__device__ Allocator2Sm() { }
|
||||
|
||||
/**
|
||||
* Performs a non-blocking allocation of TMEM.
|
||||
* @param num_columns Number of columns being freed. Must be 32 <= num_columns <= 512 and power of 2.
|
||||
* @param dst_ptr Pointer to shared memory to which to write the result tmem pointer to.
|
||||
* Both CTAs _must_ provide the exact same dst_ptr for correctness.
|
||||
* @pre Must be issued by a single fully active warp of the CTA.
|
||||
* @pre Must never be issued by more than one warp at the same time.
|
||||
* @pre For repeated allocations, the same warp must be used to issue all allocations.
|
||||
* @pre The 2 warps from participating CTAs have the same logical warp ID.
|
||||
**/
|
||||
__device__ void
|
||||
allocate(int num_columns, uint32_t* dst_ptr) {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
uint32_t dst_intptr = cute::cast_smem_ptr_to_uint(dst_ptr);
|
||||
asm volatile(
|
||||
"tcgen05.alloc.cta_group::2.sync.aligned.shared::cta.b32 [%0], %1;"
|
||||
:
|
||||
: "r"(dst_intptr), "r"(num_columns));
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
|
||||
/**
|
||||
* Frees the TMEM corresponding to the pointer and slice count provided.
|
||||
* Release the TMEM after checking that the CTA issuing the free does indeed own the corresponding slices.
|
||||
* @param tmem_ptr Base address of the TMEM address space being freed.
|
||||
* @param num_columns Number of columns being freed. Must be 32 <= num_columns <= 512 and power of 2.
|
||||
* @pre Must be issued by a single fully active warp of the CTA.
|
||||
* @pre Must never be issued by more than one warp at the same time.
|
||||
* @pre The 2 warps from participating CTAs have the same logical warp ID.
|
||||
* @returns true
|
||||
**/
|
||||
__device__
|
||||
void
|
||||
free(uint32_t tmem_ptr, int num_columns) {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
asm volatile(
|
||||
"{\n\t"
|
||||
"tcgen05.dealloc.cta_group::2.sync.aligned.b32 %0, %1; \n\t"
|
||||
"}"
|
||||
:
|
||||
: "r"(tmem_ptr), "r"(num_columns));
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
|
||||
__device__
|
||||
void
|
||||
release_allocation_lock() {
|
||||
#if defined(CUTE_ARCH_TCGEN05_TMEM_ENABLED)
|
||||
asm volatile("tcgen05.relinquish_alloc_permit.cta_group::2.sync.aligned;" ::);
|
||||
#else
|
||||
CUTE_INVALID_CONTROL_PATH("Attempting to use TMEM allocation PTX without CUTE_ARCH_TCGEN05_TMEM_ENABLED");
|
||||
#endif
|
||||
}
|
||||
};
|
||||
|
||||
} // namespace cute::TMEM
|
||||
@@ -751,14 +751,33 @@ print_latex_copy(LayoutS const& S, ThrIDS const& TS, // (m,n) -> (tid,vid) and
|
||||
#include <cute/atom/copy_traits_sm75.hpp>
|
||||
#include <cute/atom/copy_traits_sm80.hpp>
|
||||
#include <cute/atom/copy_traits_sm90.hpp>
|
||||
#include <cute/atom/copy_traits_sm100.hpp>
|
||||
|
||||
|
||||
// Config
|
||||
#if (__CUDACC_VER_MAJOR__ >= 12)
|
||||
# define CUTE_COPY_ATOM_TMA_SM90_ENABLED
|
||||
# define CUTE_COPY_ATOM_TMA_SM100_ENABLED
|
||||
#endif
|
||||
|
||||
|
||||
#if (!defined(CUTE_COPY_ATOM_TMA_SM90_ENABLED))
|
||||
# define CUTE_COPY_ATOM_TMA_SM90_ENABLED
|
||||
#endif
|
||||
|
||||
#if (!defined(CUTE_COPY_ATOM_TMA_SM100_ENABLED))
|
||||
# define CUTE_COPY_ATOM_TMA_SM100_ENABLED
|
||||
#endif
|
||||
|
||||
|
||||
#if defined(CUTE_COPY_ATOM_TMA_SM90_ENABLED)
|
||||
#include <cute/atom/copy_traits_sm90_tma.hpp>
|
||||
#endif
|
||||
|
||||
|
||||
#if defined(CUTE_COPY_ATOM_TMA_SM100_ENABLED)
|
||||
#include <cute/atom/copy_traits_sm100_tma.hpp>
|
||||
#endif
|
||||
|
||||
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
File diff suppressed because it is too large
Load Diff
@@ -0,0 +1,488 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
#pragma once
|
||||
|
||||
/*! \file
|
||||
\brief im2col make_tma_copy
|
||||
|
||||
*/
|
||||
|
||||
#include "cute/arch/copy_sm90.hpp"
|
||||
#include "cute/arch/copy_sm90_desc.hpp"
|
||||
#include "cute/atom/copy_traits_sm90_im2col.hpp"
|
||||
#include "cute/tensor.hpp"
|
||||
|
||||
namespace cute {
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_OP : SM100_TMA_2SM_LOAD_IM2COL {};
|
||||
|
||||
/// @brief Non-executable specialization of Copy_Traits for SM100
|
||||
/// im2col TMA load, with TMA descriptor but no barrier.
|
||||
///
|
||||
/// Use `.with(memory_barrier)` to construct an executable version.
|
||||
template <class NumBitsPerTMA, class TMATensor>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL, NumBitsPerTMA, TMATensor>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
Im2ColTmaDescriptor tma_desc_;
|
||||
TMATensor tma_tensor_;
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Im2ColTmaDescriptor const*
|
||||
get_tma_descriptor() const
|
||||
{
|
||||
return &tma_desc_;
|
||||
}
|
||||
|
||||
template <class GShape>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
TMATensor const
|
||||
get_tma_tensor(GShape const&) const
|
||||
{
|
||||
return tma_tensor_;
|
||||
}
|
||||
|
||||
/// @brief Get an executable specialization.
|
||||
///
|
||||
/// Copy_Traits specializations with SM100_TMA_2SM_LOAD_IM2COL are not
|
||||
/// directly executable. Instead, call this "with" member function
|
||||
/// to get an executable specialization. "Executable" means that
|
||||
/// @c copy_unpack works.
|
||||
///
|
||||
/// @param tma_mbar Memory barrier for synchronization
|
||||
///
|
||||
/// @param multicast_mask Multicast mask (unused; only exists
|
||||
/// for consistency with the actual multicast Copy_Traits
|
||||
/// specialization)
|
||||
///
|
||||
/// @return Executable specialization of @c Copy_Traits
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL_OP, NumBitsPerTMA>
|
||||
with(uint64_t& tma_mbar, [[maybe_unused]] uint16_t const& multicast_mask = 0) const
|
||||
{
|
||||
return {{}, {&tma_desc_, &tma_mbar}};
|
||||
}
|
||||
|
||||
// Copy_Traits specializations with SM100_TMA_2SM_LOAD_IM2COL
|
||||
// are not directly executable. Instead, call .with
|
||||
// to get an executable specialization.
|
||||
template <class TS, class SLayout,
|
||||
class TD, class DLayout>
|
||||
CUTE_HOST_DEVICE friend constexpr void
|
||||
copy_unpack(Copy_Traits const& traits,
|
||||
Tensor<TS,SLayout> const& src,
|
||||
Tensor<TD,DLayout> & dst) = delete;
|
||||
};
|
||||
|
||||
/// TMA load, with TMA descriptor and barrier.
|
||||
template <class NumBitsPerTMA>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL_OP, NumBitsPerTMA>
|
||||
: TMA_LOAD_IM2COL_Unpack<SM100_TMA_2SM_LOAD_IM2COL_OP>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD_IM2COL arguments
|
||||
tuple<
|
||||
Im2ColTmaDescriptor const*,
|
||||
uint64_t* // smem mbarrier
|
||||
> const opargs_;
|
||||
};
|
||||
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
///////////////////////////// TMA_LOAD_MULTICAST /////////////////////////////
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_OP : SM100_TMA_2SM_LOAD_IM2COL_MULTICAST {};
|
||||
|
||||
/// @brief Non-executable specialization of Copy_Traits for SM100
|
||||
/// im2col TMA load, with TMA descriptor but no barrier or multicast
|
||||
/// mask.
|
||||
///
|
||||
/// Use `.with(memory_barrier)` to construct an executable version.
|
||||
template <class NumBitsPerTMA, class TMATensor>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL_MULTICAST, NumBitsPerTMA, TMATensor>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
Im2ColTmaDescriptor tma_desc_;
|
||||
TMATensor tma_tensor_;
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Im2ColTmaDescriptor const*
|
||||
get_tma_descriptor() const
|
||||
{
|
||||
return &tma_desc_;
|
||||
}
|
||||
|
||||
template <class GShape>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
TMATensor const
|
||||
get_tma_tensor(GShape const&) const
|
||||
{
|
||||
return tma_tensor_;
|
||||
}
|
||||
|
||||
/// @brief Get an executable specialization.
|
||||
///
|
||||
/// Copy_Traits specializations with SM100_TMA_2SM_LOAD_IM2COL_MULTICAST
|
||||
/// are not directly executable. Instead, call this "with" member
|
||||
/// function to get an executable specialization. "Executable"
|
||||
/// means that @c copy_unpack works.
|
||||
///
|
||||
/// @param tma_mbar Memory barrier for synchronization
|
||||
///
|
||||
/// @param multicast_mask Multicast mask (defaults to a single CTA)
|
||||
///
|
||||
/// @return Executable specialization of @c Copy_Traits
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_OP, NumBitsPerTMA>
|
||||
with(uint64_t& tma_mbar, uint16_t const& multicast_mask) const
|
||||
{
|
||||
return {{}, {&tma_desc_, &tma_mbar, multicast_mask}};
|
||||
}
|
||||
|
||||
// Copy_Traits specializations with SM100_TMA_LOAD_IM2COL_MULTICAST
|
||||
// are not directly executable. Instead, call .with to get an
|
||||
// executable specialization.
|
||||
template <class TS, class SLayout,
|
||||
class TD, class DLayout>
|
||||
CUTE_HOST_DEVICE friend constexpr void
|
||||
copy_unpack(Copy_Traits const& traits,
|
||||
Tensor<TS,SLayout> const& src,
|
||||
Tensor<TD,DLayout> & dst) = delete;
|
||||
};
|
||||
|
||||
/// @brief Executable specialization of Copy_Traits for SM100 multicast
|
||||
/// im2col TMA load, with TMA descriptor, barrier, and multicast mask.
|
||||
template <class NumBitsPerTMA>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_OP, NumBitsPerTMA>
|
||||
: TMA_LOAD_IM2COL_Unpack<SM100_TMA_2SM_LOAD_IM2COL_MULTICAST_OP>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit.
|
||||
using SrcLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2, NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD_IM2COL_MULTICAST arguments
|
||||
tuple<
|
||||
Im2ColTmaDescriptor const*,
|
||||
uint64_t*, // smem mbarrier
|
||||
uint16_t // multicast mask
|
||||
> const opargs_;
|
||||
};
|
||||
|
||||
////////////////////////////////////
|
||||
// Make TMA
|
||||
///////////////////////////////////
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
/** Make a CuTe CTA-collective TiledCopy for a TMA operation.
|
||||
*
|
||||
* @param CopyOp The target copy operation: SM100_TMA_2SM_LOAD
|
||||
* @param gtensor The GMEM Tensor to be involved in the TMA.
|
||||
* @param slayout The SMEM Layout to be involved in the TMA.
|
||||
* @param cluster_tile The Cluster-local tile that each Cluster will be tiling GMEM with.
|
||||
* This is often the cluster_tile_shape that is used to tile the GMEM:
|
||||
* local_tile(gtensor, cluster_tile_shape, cluster_coord)
|
||||
* -> Cluster-local tile of GMEM
|
||||
* @param mma The TiledMMA that defines the Cluster-Tile to Block-Tile partitioning.
|
||||
*
|
||||
* This code attempts to maximize the TMA box size. It does this by tracing
|
||||
* the SMEM "vector" -- the inverse of the smem layout -- to find the largest
|
||||
* contiguous array of smem that can be written to/from global memory given
|
||||
* the constraints that the TMA instruction imposes.
|
||||
*
|
||||
* This is accomplished by assigning "basis" strides to the GMEM to track which
|
||||
* modes of SMEM map to which modes of GMEM, then reordering the modes of GMEM according
|
||||
* to the SMEM vector, and then using those GMEM/SMEM modes to fill in the desc.
|
||||
*
|
||||
* Examples:
|
||||
*/
|
||||
template <class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class Cluster_Tile,
|
||||
class... Args,
|
||||
class LowerCornerStride,
|
||||
class UpperCornerStride,
|
||||
class LowerPaddingStride,
|
||||
class UpperPaddingStride,
|
||||
class TraversalStride,
|
||||
class LowerSRTStride,
|
||||
class DilationStride>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_im2col_tma_copy_A_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (M,K,...)
|
||||
SLayout const& slayout, // (MMA, MMA_M, MMA_K)
|
||||
Cluster_Tile const& cluster_tile, // (TILE_M,TILE_N,TILE_K)
|
||||
TiledMMA<Args...> const& mma,
|
||||
LowerCornerStride const& lower_corner_whd,
|
||||
UpperCornerStride const& upper_corner_whd,
|
||||
LowerPaddingStride const& lower_padding_whd,
|
||||
UpperPaddingStride const& upper_padding_whd,
|
||||
TraversalStride const& stride_whd,
|
||||
LowerSRTStride const& lower_srt,
|
||||
DilationStride const& stride_srt,
|
||||
TMA::DescriptorAuxParams const& aux_params = {})
|
||||
{
|
||||
constexpr int R = GLayout::rank;
|
||||
// Keep only MK modes from MNK
|
||||
auto cluster_tile_shape = append<R>(make_shape(get<0>(cluster_tile), get<2>(cluster_tile)), Int<1>{});
|
||||
auto cluster_layout = make_identity_layout(cluster_tile_shape);
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_A(cluster_layout))(_, repeat<R>(_));
|
||||
|
||||
auto cta_t_vmnk_strides = [](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL_MULTICAST>) {
|
||||
return Stride<_0,_0,_1,_0>{}; // VMNK: Use only the N-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL>) {
|
||||
return Stride<_0,_0,_0,_0>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
auto cta_t_shape = shape(mma.get_thr_layout_vmnk());
|
||||
// cta rank -> logical cta idx
|
||||
auto cta_t_map = make_layout(cta_t_shape, compact_col_major(cta_t_shape, cta_t_vmnk_strides));
|
||||
|
||||
return detail::make_tma_copy_im2col(copy_op, gtensor, slayout,
|
||||
cta_t_map, cta_v_tile,
|
||||
lower_corner_whd, upper_corner_whd, lower_padding_whd, upper_padding_whd, stride_whd,
|
||||
lower_srt, stride_srt, aux_params);
|
||||
}
|
||||
|
||||
template <class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class Cluster_Tile,
|
||||
class... Args,
|
||||
class LowerCornerStride,
|
||||
class UpperCornerStride,
|
||||
class LowerPaddingStride,
|
||||
class UpperPaddingStride,
|
||||
class TraversalStride,
|
||||
class LowerSRTStride,
|
||||
class DilationStride>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_im2col_tma_copy_B_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (N,K,...)
|
||||
SLayout const& slayout, // (MMA, MMA_N, MMA_K)
|
||||
Cluster_Tile const& cluster_tile, // (TILE_M,TILE_N,TILE_K)
|
||||
TiledMMA<Args...> const& mma,
|
||||
LowerCornerStride const& lower_corner_whd,
|
||||
UpperCornerStride const& upper_corner_whd,
|
||||
LowerPaddingStride const& lower_padding_whd,
|
||||
UpperPaddingStride const& upper_padding_whd,
|
||||
TraversalStride const& stride_whd,
|
||||
LowerSRTStride const& lower_srt,
|
||||
DilationStride const& stride_srt,
|
||||
TMA::DescriptorAuxParams const& aux_params = {})
|
||||
{
|
||||
constexpr int R = GLayout::rank;
|
||||
// Keep only NK modes from MNK
|
||||
auto cluster_tile_shape = append<R>(make_shape(get<1>(cluster_tile), get<2>(cluster_tile)), Int<1>{});
|
||||
auto cluster_layout = make_identity_layout(cluster_tile_shape);
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_B(cluster_layout))(_, repeat<R>(_));
|
||||
|
||||
auto cta_t_vmnk_strides = [](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL_MULTICAST>) {
|
||||
return Stride<_0,_1,_0,_0>{}; // VMNK: Use only the M-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL>) {
|
||||
return Stride<_0,_0,_0,_0>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
auto cta_t_shape = shape(mma.get_thr_layout_vmnk());
|
||||
// cta rank -> logical cta idx
|
||||
auto cta_t_map = make_layout(cta_t_shape, compact_col_major(cta_t_shape, cta_t_vmnk_strides));
|
||||
|
||||
return detail::make_tma_copy_im2col(copy_op, gtensor, slayout,
|
||||
cta_t_map, cta_v_tile,
|
||||
lower_corner_whd, upper_corner_whd, lower_padding_whd, upper_padding_whd, stride_whd,
|
||||
lower_srt, stride_srt, aux_params);
|
||||
}
|
||||
|
||||
/////////////////////////////////////
|
||||
// Experimental Make Im2col TMA Atom
|
||||
/////////////////////////////////////
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class MMA_Tiler,
|
||||
class... Args,
|
||||
class ClusterShapeVMNK,
|
||||
class LowerCornerStride,
|
||||
class UpperCornerStride,
|
||||
class LowerPaddingStride,
|
||||
class UpperPaddingStride,
|
||||
class TraversalStride,
|
||||
class LowerSRTStride,
|
||||
class DilationStride>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_im2col_tma_atom_A_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (M, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_M, MMA_K, ...)
|
||||
MMA_Tiler const& mma_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma,
|
||||
ClusterShapeVMNK const& cluster_shape, // (CTA_V, CTA_M, CTA_N, CTA_K)
|
||||
LowerCornerStride const& lower_corner_whd,
|
||||
UpperCornerStride const& upper_corner_whd,
|
||||
LowerPaddingStride const& lower_padding_whd,
|
||||
UpperPaddingStride const& upper_padding_whd,
|
||||
TraversalStride const& stride_whd,
|
||||
LowerSRTStride const& lower_srt,
|
||||
DilationStride const& stride_srt,
|
||||
TMA::DescriptorAuxParams const& aux_params = {})
|
||||
{
|
||||
constexpr int R = GLayout::rank;
|
||||
// Keep only MK modes from MNK
|
||||
auto cluster_tile_shape = append<R>(make_shape(get<0>(mma_tiler), get<2>(mma_tiler)), Int<1>{});
|
||||
auto cluster_layout = make_identity_layout(cluster_tile_shape);
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_A(cluster_layout))(_, repeat<R>(_));
|
||||
|
||||
// The size of the multicasting
|
||||
auto num_multicast = [&](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL_MULTICAST>) {
|
||||
return size<2>(cluster_shape); // VMNK: Use only the N-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE_IM2COL> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL>) {
|
||||
return Int<1>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
return detail::make_tma_atom_im2col(copy_op, gtensor, slayout, num_multicast, cta_v_tile,
|
||||
lower_corner_whd, upper_corner_whd, lower_padding_whd, upper_padding_whd,
|
||||
stride_whd, lower_srt, stride_srt, aux_params);
|
||||
}
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class MMA_Tiler,
|
||||
class... Args,
|
||||
class ClusterShapeVMNK,
|
||||
class LowerCornerStride,
|
||||
class UpperCornerStride,
|
||||
class LowerPaddingStride,
|
||||
class UpperPaddingStride,
|
||||
class TraversalStride,
|
||||
class LowerSRTStride,
|
||||
class DilationStride>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_im2col_tma_atom_B_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (N, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_N, MMA_K, ...)
|
||||
MMA_Tiler const& mma_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma,
|
||||
ClusterShapeVMNK const& cluster_shape, // (CTA_V, CTA_M, CTA_N, CTA_K)
|
||||
LowerCornerStride const& lower_corner_whd,
|
||||
UpperCornerStride const& upper_corner_whd,
|
||||
LowerPaddingStride const& lower_padding_whd,
|
||||
UpperPaddingStride const& upper_padding_whd,
|
||||
TraversalStride const& stride_whd,
|
||||
LowerSRTStride const& lower_srt,
|
||||
DilationStride const& stride_srt,
|
||||
TMA::DescriptorAuxParams const& aux_params = {})
|
||||
{
|
||||
constexpr int R = GLayout::rank;
|
||||
// Keep only NK modes from MNK
|
||||
auto cluster_tile_shape = append<R>(make_shape(get<1>(mma_tiler), get<2>(mma_tiler)), Int<1>{});
|
||||
auto cluster_layout = make_identity_layout(cluster_tile_shape);
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_B(cluster_layout))(_, repeat<R>(_));
|
||||
|
||||
// The size of the multicasting
|
||||
auto num_multicast = [&](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL_MULTICAST>) {
|
||||
return size<1>(cluster_shape); // VMNK: Use only the M-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_IM2COL> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE_IM2COL> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_IM2COL>) {
|
||||
return Int<1>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
return detail::make_tma_atom_im2col(copy_op, gtensor, slayout, num_multicast, cta_v_tile,
|
||||
lower_corner_whd, upper_corner_whd, lower_padding_whd, upper_padding_whd,
|
||||
stride_whd, lower_srt, stride_srt, aux_params);
|
||||
}
|
||||
#endif // !defined(__CUDACC_RTC__)
|
||||
|
||||
} // end namespace cute
|
||||
@@ -0,0 +1,487 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2021 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
|
||||
|
||||
|
||||
#pragma once
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
#include <cuda.h>
|
||||
#endif
|
||||
|
||||
#include <cute/tensor.hpp>
|
||||
#include <cute/atom/copy_traits_sm90_tma.hpp>
|
||||
#include <cute/arch/copy_sm100_tma.hpp>
|
||||
#include <cute/atom/copy_traits.hpp>
|
||||
|
||||
namespace cute
|
||||
{
|
||||
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
////////////////////////////// TMA_LOAD ////////////////////////////////////////
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_OP : SM100_TMA_2SM_LOAD {};
|
||||
|
||||
// The non-executable SM100_TMA_2SM_LOAD with tma_desc and no tma_mbar
|
||||
// Use .with(tma_mbar) to construct an executable version
|
||||
template <class NumBitsPerTMA, class AuxParams_>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD, NumBitsPerTMA, AuxParams_>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD arguments
|
||||
TmaDescriptor tma_desc_;
|
||||
using AuxParams = AuxParams_;
|
||||
AuxParams aux_params_;
|
||||
|
||||
// Return TmaDescriptor/TensorMap
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
TmaDescriptor const*
|
||||
get_tma_descriptor() const {
|
||||
return &tma_desc_;
|
||||
}
|
||||
|
||||
// Construct an executable SM100_TMA_2SM_LOAD with tma_mbar
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_OP, NumBitsPerTMA>
|
||||
with(
|
||||
uint64_t& tma_mbar,
|
||||
[[maybe_unused]] uint16_t const& multicast_mask = 0,
|
||||
TMA::CacheHintSm100 const& cache_hint = TMA::CacheHintSm100::EVICT_NORMAL) const {
|
||||
// We accept multicast_mask here to keep the API for both atoms consistent
|
||||
return {{}, {&tma_desc_, &tma_mbar, static_cast<uint64_t>(cache_hint)}};
|
||||
}
|
||||
|
||||
// Construct an executable SM100_TMA_2SM_LOAD with tma_mbar (temp. overloaded for grouped gemm/ptr array gemm)
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_OP, NumBitsPerTMA>
|
||||
with(
|
||||
TmaDescriptor const* new_tma_desc,
|
||||
uint64_t& tma_mbar,
|
||||
[[maybe_unused]] uint16_t const& multicast_mask = 0,
|
||||
TMA::CacheHintSm100 const& cache_hint = TMA::CacheHintSm100::EVICT_NORMAL) const {
|
||||
// We accept multicast_mask here to keep the API for both atoms consistent
|
||||
return {{}, {new_tma_desc, &tma_mbar, static_cast<uint64_t>(cache_hint)}};
|
||||
}
|
||||
|
||||
template <class GShape>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
auto
|
||||
get_tma_tensor(GShape const& g_shape) const {
|
||||
static_assert(is_congruent<decltype(g_shape), decltype(aux_params_.g_stride_)>::value);
|
||||
return make_counting_tensor(make_layout(g_shape, aux_params_.g_stride_));
|
||||
}
|
||||
|
||||
// Don't try to execute a copy with SM100_TMA_2SM_LOAD before calling .with()
|
||||
template <class TS, class SLayout,
|
||||
class TD, class DLayout>
|
||||
CUTE_HOST_DEVICE friend constexpr void
|
||||
copy_unpack(Copy_Traits const& traits,
|
||||
Tensor<TS,SLayout> const& src,
|
||||
Tensor<TD,DLayout> & dst) = delete;
|
||||
};
|
||||
|
||||
// The executable SM100_TMA_2SM_LOAD with tma_desc and tma_mbar
|
||||
template <class NumBitsPerTMA>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_OP, NumBitsPerTMA>
|
||||
: TMA_LOAD_Unpack<SM100_TMA_2SM_LOAD_OP, NumBitsPerTMA>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD arguments
|
||||
tuple<
|
||||
TmaDescriptor const*,
|
||||
uint64_t*, // smem mbarrier
|
||||
uint64_t // cache hint
|
||||
> const opargs_;
|
||||
};
|
||||
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
///////////////////////////// TMA_LOAD_MULTICAST /////////////////////////////
|
||||
//////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
struct SM100_TMA_2SM_LOAD_MULTICAST_OP : SM100_TMA_2SM_LOAD_MULTICAST {};
|
||||
|
||||
template <class NumBitsPerTMA, class AuxParams_>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_MULTICAST, NumBitsPerTMA, AuxParams_>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD_MULTICAST_OP arguments
|
||||
TmaDescriptor tma_desc_;
|
||||
using AuxParams = AuxParams_;
|
||||
AuxParams aux_params_;
|
||||
|
||||
// Return TmaDescriptor/TensorMap
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
TmaDescriptor const*
|
||||
get_tma_descriptor() const {
|
||||
return &tma_desc_;
|
||||
}
|
||||
|
||||
// Construct an executable SM100_TMA_2SM_LOAD_MULTICAST_OP with tma_mbar
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_MULTICAST_OP, NumBitsPerTMA>
|
||||
with(
|
||||
uint64_t& tma_load_mbar,
|
||||
uint16_t const& multicast_mask,
|
||||
TMA::CacheHintSm100 const& cache_hint = TMA::CacheHintSm100::EVICT_NORMAL) const {
|
||||
return {{}, {&tma_desc_, &tma_load_mbar, multicast_mask, static_cast<uint64_t>(cache_hint)}};
|
||||
}
|
||||
|
||||
// Construct an executable SM100_TMA_2SM_LOAD_MULTICAST_OP with tma_mbar (temp. overloaded for grouped gemm/ptr array gemm)
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
Copy_Traits<SM100_TMA_2SM_LOAD_MULTICAST_OP, NumBitsPerTMA>
|
||||
with(
|
||||
TmaDescriptor const* new_tma_desc,
|
||||
uint64_t& tma_load_mbar,
|
||||
uint16_t const& multicast_mask,
|
||||
TMA::CacheHintSm100 const& cache_hint = TMA::CacheHintSm100::EVICT_NORMAL) const {
|
||||
return {{}, {new_tma_desc, &tma_load_mbar, multicast_mask, static_cast<uint64_t>(cache_hint)}};
|
||||
}
|
||||
|
||||
template <class GShape>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
auto
|
||||
get_tma_tensor(GShape const& g_shape) const {
|
||||
static_assert(is_congruent<decltype(g_shape), decltype(aux_params_.g_stride_)>::value);
|
||||
return make_counting_tensor(make_layout(g_shape, aux_params_.g_stride_));
|
||||
}
|
||||
|
||||
// Don't try to execute a copy with SM100_TMA_2SM_LOAD_MULTICAST_OP before calling .with()
|
||||
template <class TS, class SLayout,
|
||||
class TD, class DLayout>
|
||||
CUTE_HOST_DEVICE friend constexpr void
|
||||
copy_unpack(Copy_Traits const& traits,
|
||||
Tensor<TS,SLayout> const& src,
|
||||
Tensor<TD,DLayout> & dst) = delete;
|
||||
};
|
||||
|
||||
template <class NumBitsPerTMA>
|
||||
struct Copy_Traits<SM100_TMA_2SM_LOAD_MULTICAST_OP, NumBitsPerTMA>
|
||||
: TMA_LOAD_Unpack<SM100_TMA_2SM_LOAD_MULTICAST_OP, NumBitsPerTMA>
|
||||
{
|
||||
using ThrID = Layout<_2>;
|
||||
// Map from (src-thr,src-val) to bit
|
||||
using SrcLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Map from (dst-thr,dst-val) to bit
|
||||
using DstLayout = Layout<Shape<_2,NumBitsPerTMA>, Stride<NumBitsPerTMA,_1>>;
|
||||
// Reference map from (thr,val) to bit
|
||||
using RefLayout = SrcLayout;
|
||||
|
||||
// SM100_TMA_2SM_LOAD_MULTICAST_OP arguments
|
||||
tuple<
|
||||
TmaDescriptor const*,
|
||||
uint64_t*, // smem mbarrier
|
||||
uint16_t, // multicast mask
|
||||
uint64_t // cache hint
|
||||
> const opargs_;
|
||||
};
|
||||
|
||||
////////////////////////////////////
|
||||
// Make TMA
|
||||
///////////////////////////////////
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
/** Make a CuTe CTA-collective TiledCopy for a TMA operation.
|
||||
*
|
||||
* @param CopyOp The target copy operation: SM100_TMA_2SM_LOAD
|
||||
* @param gtensor The GMEM Tensor to be involved in the TMA.
|
||||
* @param slayout The SMEM Layout to be involved in the TMA.
|
||||
* @param cluster_tile The Cluster-local tile that each Cluster will be tiling GMEM with.
|
||||
* This is often the cluster_tile_shape that is used to tile the GMEM:
|
||||
* local_tile(gtensor, cluster_tile_shape, cluster_coord)
|
||||
* -> Cluster-local tile of GMEM
|
||||
* @param mma The TiledMMA that defines the Cluster-Tile to Block-Tile partitioning.
|
||||
*
|
||||
* This code attempts to maximize the TMA box size. It does this by tracing
|
||||
* the SMEM "vector" -- the inverse of the smem layout -- to find the largest
|
||||
* contiguous array of smem that can be written to/from global memory given
|
||||
* the constraints that the TMA instruction imposes.
|
||||
*
|
||||
* This is accomplished by assigning "basis" strides to the GMEM to track which
|
||||
* modes of SMEM map to which modes of GMEM, then reordering the modes of GMEM according
|
||||
* to the SMEM vector, and then using those GMEM/SMEM modes to fill in the desc.
|
||||
*
|
||||
* Examples:
|
||||
*/
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class Cluster_Tiler,
|
||||
class... Args>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_tma_copy_A_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (M, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_M, MMA_K, ...)
|
||||
Cluster_Tiler const& cluster_tiler, // (TILER_M, TILER_N, TILER_K, ...)
|
||||
TiledMMA<Args...> const& mma)
|
||||
{
|
||||
// Keep only MK modes from MNK
|
||||
auto cluster_tiler_mk = remove<1>(cluster_tiler);
|
||||
// cluster tile coord -> gtensor coord
|
||||
auto g_tile = make_identity_layout(shape(gtensor)).compose(cluster_tiler_mk); // (TILE_M, TILE_K, ...)
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_A(g_tile))(_, repeat<rank(g_tile)>(_)); // (MMA, MMA_M, MMA_K, ...)
|
||||
|
||||
auto cta_t_vmnk_strides = [](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_MULTICAST>) {
|
||||
return Stride<_0,_0,_1,_0>{}; // VMNK: Use only the N-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD>) {
|
||||
return Stride<_0,_0,_0,_0>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
auto cta_t_shape = shape(mma.get_thr_layout_vmnk());
|
||||
// cta rank -> logical cta idx
|
||||
auto cta_t_map = coalesce(make_layout(cta_t_shape, compact_col_major(cta_t_shape, cta_t_vmnk_strides)));
|
||||
|
||||
// Prefer TmaInternalType if specified. Fallback to GEngine::value_type
|
||||
using TmaType = conditional_t<is_same<void, TmaInternalType>::value, typename GEngine::value_type, TmaInternalType>;
|
||||
return detail::make_tma_copy_tiled<TmaType>(copy_op, gtensor, slayout, cta_t_map, cta_v_tile);
|
||||
}
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class Cluster_Tiler,
|
||||
class... Args>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_tma_copy_B_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (N, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_N, MMA_K, ...)
|
||||
Cluster_Tiler const& cluster_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma)
|
||||
{
|
||||
// Keep only NK modes from MNK
|
||||
auto cluster_tiler_nk = remove<0>(cluster_tiler);
|
||||
// cluster tile coord -> gtensor coord
|
||||
auto g_tile = make_identity_layout(shape(gtensor)).compose(cluster_tiler_nk); // (TILE_N, TILE_K, ...)
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_B(g_tile))(_, repeat<rank(g_tile)>(_)); // (MMA, MMA_N, MMA_K, ...)
|
||||
|
||||
auto cta_t_vmnk_strides = [](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_MULTICAST>) {
|
||||
return Stride<_0,_1,_0,_0>{}; // VMNK: Use only the M-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD>) {
|
||||
return Stride<_0,_0,_0,_0>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
auto cta_t_shape = shape(mma.get_thr_layout_vmnk());
|
||||
// cta rank -> logical cta idx
|
||||
auto cta_t_map = coalesce(make_layout(cta_t_shape, compact_col_major(cta_t_shape, cta_t_vmnk_strides)));
|
||||
|
||||
// Prefer TmaInternalType if specified. Fallback to GEngine::value_type
|
||||
using TmaType = conditional_t<is_same<void, TmaInternalType>::value, typename GEngine::value_type, TmaInternalType>;
|
||||
return detail::make_tma_copy_tiled<TmaType>(copy_op, gtensor, slayout, cta_t_map, cta_v_tile);
|
||||
}
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class Cluster_Tiler,
|
||||
class... Args>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_tma_copy_C_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (M, N, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_M, MMA_N, ...)
|
||||
Cluster_Tiler const& cluster_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma)
|
||||
{
|
||||
// Keep only MN modes from MNK
|
||||
auto cluster_tiler_mn = remove<2>(cluster_tiler);
|
||||
// cluster tile coord -> gtensor coord
|
||||
auto g_tile = make_identity_layout(shape(gtensor)).compose(cluster_tiler_mn); // (TILE_M, TILE_N, ...)
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_C(g_tile))(_, repeat<rank(g_tile)>(_)); // (MMA, MMA_M, MMA_N, ...)
|
||||
|
||||
static_assert(is_same_v<CopyOp, SM90_TMA_LOAD> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD>,
|
||||
"Unsupported TMA Op, expected a non-multicast TMA");
|
||||
|
||||
// No multicast, so only 1 CTA involved
|
||||
auto cta_t_map = Layout<_1,_0>{};
|
||||
|
||||
// Prefer TmaInternalType if specified. Fallback to GEngine::value_type
|
||||
using TmaType = conditional_t<is_same<void, TmaInternalType>::value, typename GEngine::value_type, TmaInternalType>;
|
||||
return detail::make_tma_copy_tiled<TmaType>(copy_op, gtensor, slayout, cta_t_map, cta_v_tile);
|
||||
}
|
||||
|
||||
////////////////////////////////////
|
||||
// Experimental Make TMA Atom
|
||||
///////////////////////////////////
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class MMA_Tiler,
|
||||
class... Args,
|
||||
class ClusterShapeVMNK>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_tma_atom_A_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (M, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_M, MMA_K, ...)
|
||||
MMA_Tiler const& mma_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma,
|
||||
ClusterShapeVMNK const& cluster_shape) // (CTA_V, CTA_M, CTA_N, CTA_K)
|
||||
{
|
||||
// Keep only MK modes from MNK
|
||||
auto mma_tiler_mk = remove<1>(mma_tiler);
|
||||
|
||||
// cluster tile coord -> gtensor coord
|
||||
auto g_tile = make_identity_layout(shape(gtensor)).compose(mma_tiler_mk); // (TILE_M, TILE_K, ...)
|
||||
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_A(g_tile))(_, repeat<rank(g_tile)>(_)); // (MMA, MMA_M, MMA_K, ...)
|
||||
|
||||
#if 0
|
||||
print("(tma_a) slayout: "); print(slayout); print("\n");
|
||||
print("(tma_a) mma_tiler_nk: "); print(mma_tiler_nk); print("\n");
|
||||
print("(tma_a) g_tile: "); print(g_tile); print("\n");
|
||||
print("(tma_a) mma_tiler: "); print(mma_tiler); print("\n");
|
||||
print("(tma_a) cta_v_tile: "); print(cta_v_tile); print("\n");
|
||||
#endif
|
||||
|
||||
// The size of the multicasting
|
||||
auto num_multicast = [&](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_MULTICAST>) {
|
||||
return size<2>(cluster_shape); // VMNK: Use only the N-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD>) {
|
||||
return Int<1>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
// Prefer TmaInternalType if specified. Fallback to GEngine::value_type
|
||||
using TmaType = conditional_t<is_same<void, TmaInternalType>::value, typename GEngine::value_type, TmaInternalType>;
|
||||
return detail::make_tma_copy_atom<TmaType>(copy_op, gtensor, slayout, num_multicast, cta_v_tile);
|
||||
}
|
||||
|
||||
template <class TmaInternalType = void,
|
||||
class CopyOp,
|
||||
class GEngine, class GLayout,
|
||||
class SLayout,
|
||||
class MMA_Tiler,
|
||||
class... Args,
|
||||
class ClusterShapeVMNK>
|
||||
CUTE_HOST
|
||||
auto
|
||||
make_tma_atom_B_sm100(CopyOp const& copy_op,
|
||||
Tensor<GEngine,GLayout> const& gtensor, // (N, K, ...)
|
||||
SLayout const& slayout, // (MMA, MMA_N, MMA_K, ...)
|
||||
MMA_Tiler const& mma_tiler, // (TILE_M, TILE_N, TILE_K, ...)
|
||||
TiledMMA<Args...> const& mma,
|
||||
ClusterShapeVMNK const& cluster_shape) // (CTA_V, CTA_M, CTA_N, CTA_K)
|
||||
{
|
||||
// Keep only NK modes from MNK
|
||||
auto mma_tiler_nk = remove<0>(mma_tiler);
|
||||
// cluster tile coord -> gtensor coord
|
||||
auto g_tile = make_identity_layout(shape(gtensor)).compose(mma_tiler_nk); // (TILE_N, TILE_K, ...)
|
||||
// cta val idx -> gmem mode
|
||||
auto cta_v_tile = layout<1>(mma.thrfrg_B(g_tile))(_, repeat<rank(g_tile)>(_)); // (MMA, MMA_N, MMA_K, ...)
|
||||
|
||||
#if 0
|
||||
print("(tma_b) slayout: "); print(slayout); print("\n");
|
||||
print("(tma_b) mma_tiler_nk: "); print(mma_tiler_nk); print("\n");
|
||||
print("(tma_b) g_tile: "); print(g_tile); print("\n");
|
||||
print("(tma_b) mma_tiler: "); print(mma_tiler); print("\n");
|
||||
print("(tma_b) cta_v_tile: "); print(cta_v_tile); print("\n");
|
||||
#endif
|
||||
|
||||
// The size of the multicasting
|
||||
auto num_multicast = [&](){
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD_MULTICAST> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD_MULTICAST>) {
|
||||
return size<1>(cluster_shape); // VMNK: Use only the M-CTAs in the Multicast
|
||||
} else
|
||||
if constexpr (is_same_v<CopyOp, SM90_TMA_LOAD> ||
|
||||
is_same_v<CopyOp, SM90_TMA_STORE> ||
|
||||
is_same_v<CopyOp, SM100_TMA_2SM_LOAD>) {
|
||||
return Int<1>{}; // VMNK: Use no CTAs in Non-Multicast
|
||||
} else {
|
||||
static_assert(dependent_false<CopyOp>, "Unsupported TMA");
|
||||
}
|
||||
}();
|
||||
|
||||
// Prefer TmaInternalType if specified. Fallback to GEngine::value_type
|
||||
using TmaType = conditional_t<is_same<void, TmaInternalType>::value, typename GEngine::value_type, TmaInternalType>;
|
||||
return detail::make_tma_copy_atom<TmaType>(copy_op, gtensor, slayout, num_multicast, cta_v_tile);
|
||||
}
|
||||
|
||||
#endif // !defined(__CUDACC_RTC__)
|
||||
|
||||
} // end namespace cute
|
||||
@@ -56,6 +56,13 @@ get_tma_swizzle_bits(Swizzle<B,M,S>)
|
||||
case 0: return TMA::SmemSwizzleBits::DISABLE;
|
||||
}
|
||||
} else
|
||||
|
||||
if constexpr (M == 5 || M == 6) {
|
||||
static_assert(B == 2, "Expected B = 2 when M == 5 or 6. Unsupported layout swizzle.");
|
||||
// S-condition as well?
|
||||
return TMA::SmemSwizzleBits::B128;
|
||||
} else
|
||||
|
||||
{
|
||||
static_assert(M < 0, "Unsupported layout swizzle.");
|
||||
}
|
||||
@@ -78,9 +85,25 @@ get_tma_swizzle_base(Swizzle<B,M,S>)
|
||||
static_assert(S == 3, "Expected S = 3 when M == 4. Unsupported layout swizzle.");
|
||||
return TMA::SmemSwizzleBase::SWIZZLE_BASE_16B;
|
||||
}
|
||||
|
||||
else if constexpr (M == 5) {
|
||||
static_assert(B == 2, "Expected B = 2 when M == 5. Unsupported layout swizzle.");
|
||||
static_assert(S == 2, "Expected S = 2 when M == 5. Unsupported layout swizzle.");
|
||||
return TMA::SmemSwizzleBase::SWIZZLE_BASE_32B;
|
||||
} else if constexpr (M == 6) {
|
||||
static_assert(B == 2, "Expected B = 2 when M == 5. Unsupported layout swizzle.");
|
||||
return TMA::SmemSwizzleBase::SWIZZLE_BASE_64B;
|
||||
}
|
||||
#if 1
|
||||
else {
|
||||
static_assert(4 <= M && M <= 6, "Expected 128b=16B=(2^4)B to 512b=64B=(2^6)B base swizzle.");
|
||||
}
|
||||
#else
|
||||
|
||||
else {
|
||||
static_assert(M == 4, "Expected 128b=16B=(2^4)B base swizzle.");
|
||||
}
|
||||
#endif
|
||||
}
|
||||
|
||||
template <class Layout>
|
||||
|
||||
@@ -154,6 +154,10 @@ struct MMA_Atom<MMA_Traits<MMAOperation, Args...>>
|
||||
if constexpr (has_dereference<FrgTypeA>::value) {
|
||||
// If the intended FrgTypeA is a view (of the current tensor), forward the whole
|
||||
static_assert(is_same<ValTypeA, typename remove_cvref_t<ATensor>::value_type>::value
|
||||
|
||||
|| (sizeof_bits_v<typename remove_cvref_t<ATensor>::value_type> == 8 &&
|
||||
(sizeof_bits_v<ValTypeA> == 8 || sizeof_bits_v<ValTypeA> == 6 || sizeof_bits_v<ValTypeA> == 4))
|
||||
|
||||
, "Expecting ValTypeA type");
|
||||
return make_tensor<FrgTypeA>(static_cast<ATensor&&>(atensor));
|
||||
} else {
|
||||
@@ -176,6 +180,10 @@ struct MMA_Atom<MMA_Traits<MMAOperation, Args...>>
|
||||
if constexpr (has_dereference<FrgTypeB>::value) {
|
||||
// If the intended FrgTypeB is a view (of the current tensor), forward the whole
|
||||
static_assert(is_same<ValTypeB, typename remove_cvref_t<BTensor>::value_type>::value
|
||||
|
||||
|| (sizeof_bits_v<typename remove_cvref_t<BTensor>::value_type> == 8 &&
|
||||
(sizeof_bits_v<ValTypeB> == 8 || sizeof_bits_v<ValTypeB> == 6 || sizeof_bits_v<ValTypeB> == 4))
|
||||
|
||||
, "Expecting ValTypeB type");
|
||||
return make_tensor<FrgTypeB>(static_cast<BTensor&&>(btensor));
|
||||
} else {
|
||||
@@ -1109,4 +1117,5 @@ print_svg(TiledMMA<Args...> const &mma) {
|
||||
#include <cute/atom/mma_traits_sm80.hpp>
|
||||
#include <cute/atom/mma_traits_sm90.hpp>
|
||||
#include <cute/atom/mma_traits_sm90_gmma.hpp>
|
||||
#include <cute/atom/mma_traits_sm100.hpp>
|
||||
////////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
File diff suppressed because it is too large
Load Diff
@@ -0,0 +1,109 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2023 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
|
||||
#pragma once
|
||||
|
||||
#if defined(__CUDACC_RTC__)
|
||||
#include <cuda/std/type_traits>
|
||||
#else
|
||||
#include <type_traits>
|
||||
#endif
|
||||
|
||||
#include <cute/config.hpp>
|
||||
#include <cute/tensor.hpp>
|
||||
|
||||
namespace cute {
|
||||
|
||||
//
|
||||
// A generic tiling of thread-value layouts
|
||||
//
|
||||
|
||||
template <class Layout_TV_, // (tid,vid) -> coord [Need not be 2D...]
|
||||
class Tiler_MN_> // coord space
|
||||
struct TV_Tiler
|
||||
{
|
||||
using Tiler_MN = Tiler_MN_;
|
||||
using TiledLayout_TV = Layout_TV_;
|
||||
|
||||
// Tile a tensor or a layout from shape
|
||||
// (M,N,...)
|
||||
// to shape
|
||||
// ((ThrV,FrgV),(RestM,RestN,...))
|
||||
// where
|
||||
// ThrV: The threads local to a tile.
|
||||
// FrgV: The values local to a tile.
|
||||
// RestM: The values tiled in M.
|
||||
// RestN: The values tiled in N.
|
||||
template <class Tensor>
|
||||
CUTE_HOST_DEVICE constexpr static
|
||||
auto
|
||||
apply(Tensor&& tensor)
|
||||
{
|
||||
// If Layout_TV and Tiler_MN were composable in general, then this won't be needed!
|
||||
|
||||
// ((thr_id,val_id),(RestM,RestN,...))
|
||||
return zipped_divide(tensor, Tiler_MN{}).compose(TiledLayout_TV{}, _);
|
||||
}
|
||||
|
||||
template <class SliceCoord>
|
||||
struct TV_Partitioner
|
||||
{
|
||||
SliceCoord coord_;
|
||||
|
||||
template <class TargetTensor>
|
||||
CUTE_HOST_DEVICE
|
||||
auto
|
||||
partition(TargetTensor&& target) {
|
||||
Tensor thr_tensor = make_tensor(static_cast<TargetTensor&&>(target).data(), apply(target.layout()));
|
||||
return thr_tensor(coord_, repeat<rank_v<TargetTensor>>(_));
|
||||
}
|
||||
};
|
||||
|
||||
template <class SliceCoord>
|
||||
CUTE_HOST_DEVICE static
|
||||
auto
|
||||
get_slice(SliceCoord const& coord)
|
||||
{
|
||||
return TV_Partitioner<SliceCoord>{coord};
|
||||
}
|
||||
};
|
||||
|
||||
template <class Layout_TV,
|
||||
class Tiler_MN>
|
||||
CUTE_HOST_DEVICE
|
||||
auto
|
||||
make_tiler_impl(Layout_TV const&,
|
||||
Tiler_MN const&)
|
||||
{
|
||||
return TV_Tiler<Layout_TV, Tiler_MN>{};
|
||||
}
|
||||
|
||||
}
|
||||
@@ -119,12 +119,16 @@ template <size_t N, class T>
|
||||
CUTE_HOST_DEVICE constexpr T getv(EBO<N, T, true> const&)
|
||||
{ return {}; }
|
||||
|
||||
// This is a work around approach to solve a shared memory misalign issue (https://github.com/NVIDIA/cutlass/issues/1250).
|
||||
// Will remove this work around implementation once the corresponding fix in compiler is released.
|
||||
struct dummy_EBO_base {};
|
||||
|
||||
// Specialization for types T that are not empty;
|
||||
// the "dynamic tuple leaf." Valid T here include int,
|
||||
// any other integral or floating-point type,
|
||||
// or any semiregular type for which std::is_empty_v<T> is false.
|
||||
template <size_t N, class T>
|
||||
struct EBO<N, T, false>
|
||||
struct EBO<N, T, false> : private dummy_EBO_base
|
||||
{
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
EBO() : t_{} {}
|
||||
|
||||
@@ -78,6 +78,7 @@ using int_byte_t = typename int_byte<N>::type;
|
||||
using uint1_t = cutlass::uint1b_t;
|
||||
using uint2_t = cutlass::uint2b_t;
|
||||
using uint4_t = cutlass::uint4b_t;
|
||||
using uint6_t = cutlass::uint6b_t;
|
||||
using CUTE_STL_NAMESPACE::uint8_t;
|
||||
using CUTE_STL_NAMESPACE::uint16_t;
|
||||
using CUTE_STL_NAMESPACE::uint32_t;
|
||||
@@ -88,6 +89,7 @@ template <int N> struct uint_bit;
|
||||
template <> struct uint_bit< 1> { using type = uint1_t; };
|
||||
template <> struct uint_bit< 2> { using type = uint2_t; };
|
||||
template <> struct uint_bit< 4> { using type = uint4_t; };
|
||||
template <> struct uint_bit< 6> { using type = uint6_t; };
|
||||
template <> struct uint_bit< 8> { using type = uint8_t; };
|
||||
template <> struct uint_bit< 16> { using type = uint16_t; };
|
||||
template <> struct uint_bit< 32> { using type = uint32_t; };
|
||||
|
||||
@@ -73,6 +73,29 @@ using cutlass::uint4b_t;
|
||||
using cutlass::bin1_t;
|
||||
|
||||
|
||||
using cutlass::float_ue4m3_t;
|
||||
using cutlass::float_ue8m0_t;
|
||||
|
||||
using cutlass::uint6b_t;
|
||||
using cutlass::float_e2m1_t;
|
||||
using cutlass::float_e2m3_t;
|
||||
using cutlass::float_e3m2_t;
|
||||
|
||||
using cutlass::type_erased_dynamic_float6_t;
|
||||
using cutlass::type_erased_dynamic_float4_t;
|
||||
|
||||
namespace detail {
|
||||
using cutlass::detail::float_e2m1_unpacksmem_t;
|
||||
using cutlass::detail::float_e2m3_unpacksmem_t;
|
||||
using cutlass::detail::float_e3m2_unpacksmem_t;
|
||||
using cutlass::detail::float_e2m3_unpack8bits_t;
|
||||
using cutlass::detail::float_e3m2_unpack8bits_t;
|
||||
using cutlass::detail::type_erased_dynamic_float4_unpacksmem_t;
|
||||
using cutlass::detail::type_erased_dynamic_float6_unpacksmem_t;
|
||||
};
|
||||
|
||||
|
||||
|
||||
//
|
||||
// Print utility
|
||||
//
|
||||
@@ -133,4 +156,26 @@ pretty_print(float_e5m2_t t) {
|
||||
printf("%*.2f", 8, static_cast<float>(t));
|
||||
}
|
||||
|
||||
|
||||
template <
|
||||
cutlass::detail::FpEncoding Encoding,
|
||||
class Derived
|
||||
>
|
||||
CUTE_HOST_DEVICE
|
||||
void
|
||||
print(cutlass::float_exmy_base<Encoding, Derived> a) {
|
||||
printf("%f", static_cast<float>(a));
|
||||
}
|
||||
|
||||
template <
|
||||
cutlass::detail::FpEncoding Encoding,
|
||||
class Derived
|
||||
>
|
||||
CUTE_HOST_DEVICE
|
||||
void
|
||||
pretty_print_float_exmy_base(cutlass::float_exmy_base<Encoding, Derived> t) {
|
||||
printf("%*.2f", 8, static_cast<float>(t));
|
||||
}
|
||||
|
||||
|
||||
} // namespace cute
|
||||
|
||||
@@ -284,6 +284,96 @@ recast_ptr(rmem_ptr<P> const& ptr) {
|
||||
return make_rmem_ptr(recast_ptr<NewT>(ptr.get()));
|
||||
}
|
||||
|
||||
|
||||
//
|
||||
// tmem_ptr -- a typed, word-addressed, non-dereferencable "pointer"
|
||||
//
|
||||
|
||||
template <class T>
|
||||
struct tmem_ptr
|
||||
{
|
||||
using value_type = remove_cv_t<T>;
|
||||
using element_type = T;
|
||||
using reference = T;
|
||||
|
||||
// Right-shift value for the offset scaling -- TMEM uses word-addressing
|
||||
static constexpr int32_t OffsetShift = log_2(trait_ratio(sizeof_bits<uint32_t>{}, sizeof_bits<T>{}));
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
tmem_ptr(uint32_t addr = 0) : addr_(addr) {}
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
uint32_t const& get() const {
|
||||
return addr_;
|
||||
}
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
uint32_t& get() {
|
||||
return addr_;
|
||||
}
|
||||
|
||||
template <class T_ = T>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
value_type operator*() const {
|
||||
static_assert(dependent_false<T_>, "Attempting to dereference a tmem_ptr, want raw_pointer_cast() for address instead?");
|
||||
return value_type{};
|
||||
}
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
reference operator[](uint32_t const& i) const { return *(*this + i); }
|
||||
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
tmem_ptr operator+(uint32_t const& i) const {
|
||||
//return {addr_ + shiftr(i, OffsetShift)}; // Shift the offset for word-addressing
|
||||
return {addr_ + rotr(i, OffsetShift)}; // Rotate the offset to keep subword indices in the unused high 8bits for debug
|
||||
}
|
||||
|
||||
// TMEM "Address" with active mask 0x007F.01FF
|
||||
// The upper 16 bits, the 0x007F portion, refers to the 128 DP lanes
|
||||
// The lower 16 bits, the 0x01FF portion, refers to the 512 COL lanes
|
||||
union {
|
||||
uint32_t addr_;
|
||||
struct {
|
||||
uint16_t col_;
|
||||
uint8_t dp_;
|
||||
uint8_t idx_; // Hijack the top 8bits for the sub-word idx to avoid an extra reg.
|
||||
// Assert this is 0 on every access?
|
||||
};
|
||||
};
|
||||
};
|
||||
|
||||
template <class T, class = void>
|
||||
struct is_tmem : false_type {};
|
||||
template <class T> // Found the tmem
|
||||
struct is_tmem<tmem_ptr<T>> : true_type {};
|
||||
template <class P> // Recurse on ::iterator, if possible
|
||||
struct is_tmem<P, void_t<typename P::iterator>> : is_tmem<typename P::iterator> {};
|
||||
template <class P>
|
||||
constexpr bool is_tmem_v = is_tmem<P>::value;
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
tmem_ptr<T>
|
||||
make_tmem_ptr(uint32_t addr = 0) {
|
||||
return tmem_ptr<T>(addr);
|
||||
}
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
uint32_t
|
||||
raw_pointer_cast(tmem_ptr<T> const& ptr) {
|
||||
return ptr.get();
|
||||
}
|
||||
|
||||
// TMEM accounts for subword/superword elements already due to the offset shift based on sizeof_bits
|
||||
// Thus, this is a trivial recast equivalent to reinterpret_cast<NewT*>
|
||||
template <class NewT, class T>
|
||||
CUTE_HOST_DEVICE constexpr
|
||||
auto
|
||||
recast_ptr(tmem_ptr<T> const& ptr) {
|
||||
return tmem_ptr<NewT>{ptr.addr_};
|
||||
}
|
||||
|
||||
|
||||
//
|
||||
// Display utilities
|
||||
//
|
||||
@@ -306,6 +396,14 @@ CUTE_HOST_DEVICE void print(rmem_ptr<T> ptr)
|
||||
printf("rmem_"); print(ptr.get());
|
||||
}
|
||||
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE void print(tmem_ptr<T> ptr)
|
||||
{
|
||||
printf("tmem_["); print(sizeof_bits<T>::value); printf("b](0x%04x.%04x)", ptr.addr_ >> 16, ptr.addr_ & 0xFFFF);
|
||||
}
|
||||
|
||||
|
||||
#if !defined(__CUDACC_RTC__)
|
||||
template <class T>
|
||||
CUTE_HOST std::ostream& operator<<(std::ostream& os, gmem_ptr<T> ptr)
|
||||
@@ -325,6 +423,13 @@ CUTE_HOST std::ostream& operator<<(std::ostream& os, rmem_ptr<T> ptr)
|
||||
return os << "rmem_[" << int(sizeof_bits<iter_value_t<T>>::value) << "b]";
|
||||
}
|
||||
|
||||
|
||||
template <class T>
|
||||
CUTE_HOST std::ostream& operator<<(std::ostream& os, tmem_ptr<T> ptr)
|
||||
{
|
||||
return os << "tmem_[" << int(sizeof_bits<T>::value) << "b](" << ptr.addr_ << ")";
|
||||
}
|
||||
|
||||
#endif // !defined(__CUDACC_RTC__)
|
||||
|
||||
} // end namespace cute
|
||||
|
||||
@@ -95,6 +95,9 @@ template <class... Iters>
|
||||
struct is_smem<ZipIterator<Iters...>> : conjunction<is_smem<Iters>...> {};
|
||||
template <class... Iters>
|
||||
struct is_gmem<ZipIterator<Iters...>> : conjunction<is_gmem<Iters>...> {};
|
||||
template <class... Iters>
|
||||
struct is_tmem<ZipIterator<Iters...>> : conjunction<is_tmem<Iters>...> {};
|
||||
|
||||
// A tuple of Layouts that operates on each Layout symmetrically
|
||||
// The Layouts need to have compatible shapes and ranks.
|
||||
// The ZipLayout presents the intersection of the domain of its component Layouts.
|
||||
|
||||
@@ -255,7 +255,12 @@ pretty_print(double v) {
|
||||
template <class T>
|
||||
CUTE_HOST_DEVICE void
|
||||
pretty_print(T t) {
|
||||
constexpr auto has_print_exmy_base = cute::is_valid([](auto t) -> decltype(pretty_print_float_exmy_base(t)) {}, t);
|
||||
if constexpr (has_print_exmy_base) {
|
||||
pretty_print_float_exmy_base(t);
|
||||
} else {
|
||||
printf(" "); print(t);
|
||||
}
|
||||
}
|
||||
|
||||
} // end namespace cute
|
||||
|
||||
Reference in New Issue
Block a user